"""
Description: Elbow connection between two elements.
"""
# Copyright (c) 2024 - 2025 Milan Piskla
# Licensed under the MIT License - see LICENSE file for details

from ..base.AbstractElement import AbstractElement
from .AaLineSegmentIntersection import AaLineSegmentIntersection
from .Directions import Directions

from .Point import Point
from typing import Optional, Callable

import copy

MIN_SEGMENT_HINT: float = 30.0
MAX_ELBOWS = 8

Segment = tuple[Point, Point]

class ElbowConnection:
    class Trajectory:
        def __init__(self):
            self._points: list[Point] = []
            self._distance: float = 0.0

        def is_better_than(self, other: 'ElbowConnection.Trajectory') -> bool:
            if len(other._points) == 0:
                return True

            if len(self._points) < len(other._points):
                return True
            
            if len(self._points) == len(other._points):
                if self._distance < other._distance:
                    return True
                
            return False

        def get_elbows_count(self) -> int:
            return len(self._points)

        def get_points(self) -> list[Point]:
            return self._points
        
        def add_point(self, point: Point) -> None:
            if len(self._points) > 0:
                xlast, ylast = self._points[-1]
                x, y = point
                self._distance += max(abs(x - xlast), abs(y - ylast))
            
            self._points.append(point)

        def pop_point(self):
            self._points.pop()

    def __init__(self, start_element: AbstractElement, end_element: AbstractElement, start_dir: str, end_dir: str, min_segment_hint = MIN_SEGMENT_HINT):
        self._start_element = start_element
        self._end_element = end_element
        self._start_dir = start_dir
        self._end_dir = end_dir

        self._start_point = self._find_edge_point(self._start_element, self._start_dir)
        self._end_point = self._find_edge_point(self._end_element, self._end_dir)

        self._vertical_segments: list[Segment] = []
        self._horizontal_segments: list[Segment] = []

        self._min_segment_hint = min_segment_hint

        self._xmin = 0.0
        self._xmax = 0.0
        self._ymin = 0.0
        self._ymax = 0.0

        self._best_trajectory = ElbowConnection.Trajectory()
        self._current_trajectory = ElbowConnection.Trajectory()

        self._is_already_computed = False

    def points(self) -> list[Point]:
        if not self._is_already_computed:
            self._compute_bounds()
            start_segment = self._find_free_segment(self._start_point, self._start_dir, self._end_element)
            if not start_segment or not AaLineSegmentIntersection.is_aa_segment(start_segment[0], start_segment[1]):
                return [self._start_point, self._end_point]
        
            self._fill_segments()
                
            p1, p2 = start_segment
            self._current_trajectory.add_point(p1)
            self._cross_vertical(p1, p2) if AaLineSegmentIntersection.is_horizontal_segment(p1, p2) else self._cross_horizontal(p1, p2)

            self._is_already_computed = True

        return self._best_trajectory.get_points()

    def _find_edge_point(self, element: AbstractElement, direction: str) -> Point:
        if direction not in Directions.keys():
            raise ValueError(f'Wrong direction {direction}, shoould be one of the U, L, D, R')
        
        cx, cy = element.center()
        a, b = element._width / 2, element._height / 2
        dx, dy = Directions.dxdy(direction)
        
        return cx + a * dx, cy + b * dy

    def _find_horizontal_gap_segment(self) -> Optional[Segment]:
        """Find a line in the middle of the horizontal space between the rectangles."""
        y1 = max(self._start_element._y, self._end_element._y)
        y2 = min(self._start_element._y + self._start_element._height, self._end_element._y + self._end_element._height)

        return ((self._xmin, (y1 + y2) / 2), (self._xmax, (y1 + y2) / 2)) if y1 > y2 else None
    
    def _find_vertical_gap_segment(self) -> Optional[Segment]:
        """Find a line in the middle of the vertical space between the rectangles."""
        x1 = max(self._start_element._x, self._end_element._x)
        x2 = min(self._start_element._x + self._start_element._width, self._end_element._x + self._end_element._width)
        
        return (((x1 + x2) / 2, self._ymin), ((x1 + x2) / 2, self._ymax)) if x1 > x2 else None

    def _compute_bounds(self) -> None:
        self._xmin = min(self._start_element._x, self._end_element._x) - self._min_segment_hint
        self._xmax = max(self._start_element._x + self._start_element._width, self._end_element._x + self._end_element._width) + self._min_segment_hint
        self._ymin = min(self._start_element._y, self._end_element._y) - self._min_segment_hint
        self._ymax = max(self._start_element._y + self._start_element._height, self._end_element._y + self._end_element._height) + self._min_segment_hint

    def _fill_segments(self) -> None:
        self._add_segment_if_valid(self._find_horizontal_gap_segment())
        self._add_segment_if_valid(self._find_vertical_gap_segment())
        self._add_segment_if_valid(self._find_free_segment(self._end_point, self._end_dir, self._start_element))
        self._add_segment(((self._xmin, self._ymin), (self._xmax, self._ymin)))
        self._add_segment(((self._xmin, self._ymax), (self._xmax, self._ymax)))
        self._add_segment(((self._xmin, self._ymin), (self._xmin, self._ymax)))
        self._add_segment(((self._xmax, self._ymin), (self._xmax, self._ymax)))

    def _add_segment_if_valid(self, segment: Optional[Segment]) -> None:
        if segment:
            self._add_segment(segment)
            
    def _add_segment(self, segment: Segment) -> None:
        if AaLineSegmentIntersection.is_vertical_segment(segment[0], segment[1]):
            self._vertical_segments.append(segment)
        elif AaLineSegmentIntersection.is_horizontal_segment(segment[0], segment[1]):
            self._horizontal_segments.append(segment)
        else:
            raise Exception('Not AA segment')

    def _find_free_segment(self, start: Point, direction: str, obstacle: AbstractElement) -> Optional[Segment]:
        def manhattan_distance(p1: Point, p2: Point) -> float:
            return abs(p1[0] - p2[0]) + abs(p1[1] - p2[1])

        x, y = start
        w, h = self._xmax - self._xmin, self._ymax - self._ymin
        dx, dy = Directions.dxdy(direction)
        end = self._clamp_point((x + w * dx, y + h * dy))

        intersections = AaLineSegmentIntersection.with_rectangle(start, end, obstacle.center(), obstacle._width / 2, obstacle._height / 2)
        return (start, end) if not intersections else (start, min(intersections, key = lambda point: manhattan_distance(point, start)))

    def _clamp_point(self, point: Point) -> Point:
        x, y = point
        return (min(max(x, self._xmin), self._xmax), min(max(y, self._ymin), self._ymax))

    # Recursive magic follows :)
    # TODO optimize the algorithm to not cross segments that were already 
    # intersected in previous iterations

    def _cross_horizontal(self, p1: Point, p2: Point) -> None:
        self._cross_segment(p1, p2, self._cross_vertical, self._horizontal_segments)

    def _cross_vertical(self, p1: Point, p2: Point) -> None:
        self._cross_segment(p1, p2, self._cross_horizontal, self._vertical_segments)

    def _cross_segment(self, p1: Point, p2: Point, cross_function: Callable, cross_segments: list[Segment]) -> None:
        if self._current_trajectory.get_elbows_count() < MAX_ELBOWS:
            self._try_complete_trajectory(p1, p2)
            for q1, q2 in cross_segments:
                intersection = AaLineSegmentIntersection.with_aa_line_segment(p1, p2, q1, q2)
                if intersection:
                    self._current_trajectory.add_point(intersection)
                    cross_function(q1, q2)
                    self._current_trajectory.pop_point()

    def _try_complete_trajectory(self, p1: Point, p2: Point) -> None:
        if AaLineSegmentIntersection.is_point_on_segment(self._end_point, p1, p2):
            self._current_trajectory.add_point(self._end_point)
            if self._current_trajectory.is_better_than(self._best_trajectory):
                self._best_trajectory = copy.deepcopy(self._current_trajectory)
            
            self._current_trajectory.pop_point()



"""
Description: Generates indexes for Excalidraw elements.
Excalidraw generates fractional indexes according to 
https://observablehq.com/@dgreensp/implementing-fractional-indexing

We just need to generate a sequece of fractional indexes for the fixed amount 
of elements in the diagram.
"""
# Copyright (c) 2024 - 2025 Milan Piskla
# Licensed under the MIT License - see LICENSE file for details

class IndexGenerator:
    _BASE62_CHARS = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"

    def __init__(self, initial_value='a0'):
        """
        Initialize the generator with the given initial value.
        
        :param initial_value: The starting value of the index, default is 'Zz' to get first next value 'a0'.
        """
        self._current = initial_value

    def next(self):
        """
        Generate the next index in the sequence.

        :return: The next fractional index as a string.
        """
        # Increment the current value
        self._current = self._increment(self._current)

        # Handle transitions like 'az' -> 'b00'
        if all(c == self._BASE62_CHARS[0] for c in self._current[1:]):
            self._current += self._BASE62_CHARS[0]

        return self._current


    def _increment(self, value):
        """
        Increment a Base62 string value with proper carry-over.

        :param value: The Base62 string to increment.
        :return: The incremented Base62 string.
        """
        chars = list(value)
        i = len(chars) - 1
        carry = True

        while i >= 0 and carry:
            char_index = self._BASE62_CHARS.index(chars[i]) + 1
            if char_index < len(self._BASE62_CHARS):
                chars[i] = self._BASE62_CHARS[char_index]
                carry = False
            else:
                chars[i] = self._BASE62_CHARS[0]
                i -= 1

        if carry:
            chars.insert(0, self._BASE62_CHARS[0])

        return ''.join(chars)


"""
Description: Text element.
"""
# Copyright (c) 2024 - 2025 Milan Piskla
# Licensed under the MIT License - see LICENSE file for details

from ..base.AbstractElement import AbstractElement
from ..colors.Color import Color
from ...defaults.Defaults import Defaults
from ..inputs.Font import Font
from ..inputs.Fontsize import Fontsize
from ..inputs.Align import Align
from ..inputs.Baseline import Baseline
from typing import Self, overload

_ANCHOR_OFFSETS_COEFFS = {
    "left" : {
        "top" : (0.0, 0.0),
        "middle" : (0.0, 0.5),
        "bottom" : (0.0, 1.0) 
    },
    "center" : {
        "top" : (0.5, 0.0),
        "middle" : (0.5, 0.5),
        "bottom" : (0.5, 1.0) 
    },
    "right" : {
        "top" : (1.0, 0.0),
        "middle" : (1.0, 0.5),
        "bottom" : (1.0, 1.0) 
    }
}

class Text(AbstractElement):
    """A class representing text elements in excaligen.
    The Text class is designed to handle and manipulate text elements within the excaligen 
    framework. It provides various text customization options including font styles, sizes, 
    alignments, and colors. The class supports auto-resizing capabilities and multiple font 
    families.

    > [!WARNING]
    > Do not instantiate this class directly. Use `SceneBuilder.text()` instead.
    """
    CHAR_WIDTH_FACTOR = 0.6  # Approximate width of a character relative to the font size
    LINE_HEIGHT_FACTOR = 1.25  # Approximate line height factor

    def __init__(self, defaults: Defaults, text: str | None = None):
        super().__init__("text", defaults)
        self._text: str = text if text is not None else ""
        self._font_size = getattr(defaults, "_font_size")
        self._font_family = getattr(defaults, "_font_family")
        self._text_align = getattr(defaults, "_text_align")
        self._vertical_align = getattr(defaults, "_vertical_align")
        self._line_height = getattr(defaults, "_line_height")
        self._auto_resize = getattr(defaults, "_auto_resize")
        self._stroke_color = getattr(defaults, "_stroke_color")
        self._width = getattr(defaults, "_width")
        self._height = getattr(defaults, "_height")
        self._container_id: str | None = None
        self.__is_anchored: bool = False
        self.__calculate_dimensions()

    def content(self, text: str) -> Self:
        """Set the text content and automatically calculate width and height.

        Args:
            text (str): The text content to set.

        Returns:
            Self: The current instance of the Text class.
        """
        def action():
            self._text = text
            self.__calculate_dimensions()

        self.__check_anchor_and_do_action(action)
        return self

    def fontsize(self, size: int | str) -> Self:
        """Set the font size by int or by string ('S', 'M', 'L', 'XL').

        Args:
            size (int | str): The font size to set.

        Raises:
            ValueError: If an invalid size string is provided.
            TypeError: If the size is not an int or a valid string.

        Returns:
            Self: The current instance of the Text class.
        """
        def action():
            self._font_size = Fontsize.from_(size)
            self.__calculate_dimensions()

        self.__check_anchor_and_do_action(action)
        return self

    def font(self, family: str) -> Self:
        """Set the font family ('Excalifont', 'Comic Shaans', 'Lilita One', 'Nunito', 'Hand-drawn', 'Normal', 'Code').

        Args:
            family (str): The font family to set.

        Raises:
            ValueError: If an invalid font family is provided.

        Returns:
            Self: The current instance of the Text class.
        """
        self._font_family = Font.from_(family)
        return self

    def align(self, align: str) -> Self:
        """Set the horizontal text alignment (left, center, right).

        Args:
            align (str): The horizontal alignment to set.

        Raises:
            ValueError: If an invalid alignment is provided.

        Returns:
            Self: The current instance of the Text class.
        """
        def action():
            self._text_align = Align.from_(align)
            self.__calculate_dimensions()

        self.__check_anchor_and_do_action(action)
        return self

    def baseline(self, baseline: str) -> Self:
        """Set the vertical text alignment (top, middle, bottom).

        Args:
            baseline (str): The vertical alignment to set.

        Raises:
            ValueError: If an invalid vertical alignment is provided.

        Returns:
            Self: The current instance of the Text class.
        """
        def action():
            self._vertical_align = Baseline.from_(baseline)
            self.__calculate_dimensions()

        self.__check_anchor_and_do_action(action)
        return self

    def spacing(self, height: float) -> Self:
        """Set the line height manually.

        Args:
            height (float): The line height to set.

        Returns:
            Self: The current instance of the Text class.
        """
        self._line_height = height
        return self

    def autoresize(self, enabled: bool) -> Self:
        """Enable or disable automatic text box resizing.

        Args:
            enabled (bool): Whether to enable automatic resizing.

        Returns:
            Self: The current instance of the Text class.
        """
        self._auto_resize = enabled
        return self

    def color(self, color: str | Color) -> Self:
        """Set the text color as #RRGGBB, color name or Color object.

        Args:
            color (str | Color): The color to set.

        Returns:
            Self: The current instance of the Text class.
        """
        self._stroke_color = Color.from_(color)
        return self
    
    def anchor(self, x: float, y: float, align: str | None = None, baseline: str | None = None) -> Self:
        """Anchor the text element to a specific point (x, y).
        It takes horizontal and vertical alignment into account.

        Args:
            x (float): The x-coordinate to anchor to.  
            y (float): The y-coordinate to anchor to.  
            align (str | None, optional): The horizontal alignment ('left', 'center', 'right').  
            baseline (str | None, optional): The vertical alignment ('top', 'middle', 'bottom').

        Returns:
            Self: The current instance of the Text class.
        """
        self.__is_anchored = True

        if align:
            self.align(align)

        if baseline:
            self.baseline(baseline)
        
        self.__do_anchor(x, y)
        return self
    
    @overload
    def center(self) -> tuple[float, float]: ...

    @overload
    def center(self, x: float, y: float) -> Self: ...

    def center(self, *args) -> Self | tuple[float, float]:
        """
        Get or set the center coordinates of the element.

        Args:
            *args: Supports two signatures:
                1. center() -> tuple[float, float]
                    Returns the (x, y) coordinates of the center.
                2. center(x, y) -> Self
                    Sets the center to (x, y) and returns self for chaining.

        Returns:
            tuple[float, float] | Self: Depending on the arguments.
        """
        match args:
            case ():
                return super().center()
            case (x, y):
                return self.anchor(x, y, "center", "middle")
            case _:
                raise ValueError("Invalid arguments for center. Expected () or (x, y).")
    
    def justify(self, x: float, y: float, width: float, height: float) -> Self:
        """Justify the text element within a rectangle defined by (x, y, width, height).

        Args:
            x (float): The x-coordinate of the rectangle's top-left corner.
            y (float): The y-coordinate of the rectangle's top-left corner.
            width (float): The width of the rectangle.
            height (float): The height of the rectangle.

        Returns:
            Self: The current instance of the Text class.
        """
        self.__is_anchored = True
        cx, cy = _ANCHOR_OFFSETS_COEFFS[self._text_align][self._vertical_align]
        self.__do_anchor(x + cx * width, y + cy * height)
        return self

    def __calculate_dimensions(self):
        """Calculate the width and height based on the text content."""
        lines = self._text.split("\n")
        width = max(len(line) for line in lines) * self._font_size * self.CHAR_WIDTH_FACTOR
        height = len(lines) * self._font_size * self.LINE_HEIGHT_FACTOR

        self._size(width, height)

    def __do_anchor(self, x: float, y: float) -> None:
        """Calculate the position based on anchoring and alignment."""
        cx, cy = _ANCHOR_OFFSETS_COEFFS[self._text_align][self._vertical_align]
        self._x = x - cx * self._width
        self._y = y - cy * self._height

    def __get_anchor(self) -> tuple[float, float]:
        """Get the anchor point based on current position and alignment.

        Returns:
            tuple[float, float]: The (x, y) coordinates of the anchor point.
        """
        cx, cy = _ANCHOR_OFFSETS_COEFFS[self._text_align][self._vertical_align]
        return (self._x + cx * self._width, self._y + cy * self._height)
    
    def __check_anchor_and_do_action(self, action) -> None:
        """Check if anchored, get anchor point, do action, re-anchor."""
        if self.__is_anchored:
            x, y = self.__get_anchor()
            action()
            self.__do_anchor(x, y)
        else:
            action()



"""
Description: Straight connection between two elements.
"""
# Copyright (c) 2024 - 2026 Milan Piskla
# Licensed under the MIT License - see LICENSE file for details

from ..base.AbstractElement import AbstractElement
from .HalfLineIntersection import HalfLineIntersection
from .Point import Point

from typing import Optional

class StraightConnection:
    def __init__(self, start_element: AbstractElement, end_element: AbstractElement):
        self._start_element = start_element
        self._end_element = end_element

    def points(self) -> list[Point]:
        xs, ys = self._start_element.center()
        xe, ye = self._end_element.center()

        points = []

        start_point = self.__find_intersection_with_element(self._start_element, 0.0, 0.0, xe - xs, ye - ys)
        if start_point is None:
            raise Exception("Cannot find intersection as start point")
        points.append(start_point)

        end_point = self.__find_intersection_with_element(self._end_element, 0.0, 0.0, xs - xe, ys - ye)
        if end_point is None:
            raise Exception("Cannot find intersection as end point")
        points.append(end_point)
        
        return points

    def __find_intersection_with_element(self, element: AbstractElement, dx: float, dy: float, vx: float, vy: float) -> Optional[Point]:
        """Find the precise intersection point between a half-line and the element's edge."""

        a, b = element._width / 2, element._height / 2
        intersection: Point = (0.0, 0.0)

        match element._type:
            case "rectangle" | "image" | "text":
                intersection = HalfLineIntersection.with_rectangle(dx, dy, vx, vy, a, b, element._angle) # type: ignore
            
            case "diamond":
                intersection = HalfLineIntersection.with_diamond(dx, dy, vx, vy, a, b, element._angle) # type: ignore
            
            case "ellipse":
                intersection = HalfLineIntersection.with_ellipse(dx, dy, vx, vy, a, b, element._angle) # type: ignore

            case _:
                raise TypeError(f"Cannot find intersection with unknown type {element._type}")
            
        if intersection is None:
            return None

        return tuple(p + t for p, t in zip(intersection, element.center()))
            



"""
Description: Functional tests for arrows.
"""
# Copyright (c) 2024 - 2026 Milan Piskla
# Licensed under the MIT License - see LICENSE file for details

from excaligen.SceneBuilder import SceneBuilder
from .evaluate import *
from typing import Any

from pytest import FixtureRequest
import math

def test_arrow_star(reference_json: dict[str, Any], request: FixtureRequest) -> None:
    xg = SceneBuilder()

    start_element = xg.ellipse('center').center(0, 0).size(100, 100)

    for angle in range(0, 360, 30):
        radians = angle * math.pi / 180
        x = 300 * math.cos(radians)
        y = 300 * math.sin(radians)
        end_element = xg.rectangle(str(angle)).center(x, y).size(100, 50).roundness('round')
        xg.arrow().bind(start_element, end_element)

    evaluate(reference_json, xg, request)

def test_arrow_curve(reference_json: dict[str, Any], request: FixtureRequest) -> None:
    scene = SceneBuilder()
    
    center = scene.rectangle('center').center(0, 0).size(160, 70)
    ur = scene.ellipse('up-right').center(400, -200).size(130, 50)
    ul = scene.ellipse('up-left').center(-400, -200).size(130, 50)
    dr = scene.ellipse('down-right').center(200, 300).size(130, 50)
    dl = scene.ellipse('down-left').center(-200, 300).size(130, 50)
   
    scene.arrow().curve('R', 'L').bind(center, ur).arrowheads(None, 'arrow')
    scene.arrow().curve('L', 'R').bind(center, ul).arrowheads(None, 'arrow')
    scene.arrow().curve('D', 'U').bind(center, dr).arrowheads(None, 'arrow')
    scene.arrow().curve('D', 'U').bind(center, dl).arrowheads(None, 'arrow')
    
    evaluate(reference_json, scene, request)

def test_arrow_curve_narrow(reference_json: dict[str, Any], request: FixtureRequest) -> None:
    xg = SceneBuilder()
    start_element = xg.ellipse('Begin').center(-150, -150).size(120, 80)
    end_element = xg.ellipse('End').center(150, 150).size(120, 80)
    xg.arrow().curve(0.0, 3.14 * 0.75).bind(start_element, end_element)

    evaluate(reference_json, xg, request)

def test_arrow_curve_complex(reference_json: dict[str, Any], request: FixtureRequest) -> None:
    xg = SceneBuilder()
    RADIUS = 700
    XOFFSET = -200
    center_element = xg.rectangle('center').center(0, 0).size(120, 80).roundness('round')

    for i in range(-3, 4):
        angle = i * math.pi / 14
        xr = RADIUS * math.cos(angle) + XOFFSET
        yr = RADIUS * math.sin(angle)
        relement = xg.rectangle(f'right {i}').center(xr, yr).size(120, 80).roundness('round')
        xg.arrow().curve(0, math.pi).bind(center_element, relement).arrowheads(None, 'arrow')
        
        xl = RADIUS * math.cos(angle + math.pi) - XOFFSET
        yl = RADIUS * math.sin(angle + math.pi)
        lelement = xg.rectangle(f'left {i}').center(xl, yl).size(120, 80).roundness('round')
        xg.arrow().curve(math.pi, 0).bind(center_element, lelement).arrowheads(None, 'arrow')

    evaluate(reference_json, xg, request)

def test_arrow_beams_as_curves(reference_json: dict[str, Any], request: FixtureRequest) -> None:
    scene = SceneBuilder()
    start_element = scene.ellipse('Center').center(0, 0).size(100, 100)
    
    for angle in range(0, 360, 30):
        radians = angle * math.pi / 180
        x = 300 * math.cos(radians)
        y = 300 * math.sin(radians)
        end_element = scene.ellipse(str(angle)).center(x, y).size(80, 80)
        scene.arrow().curve(radians, radians -math.pi).bind(start_element, end_element)

    evaluate(reference_json, scene, request)

def test_arrow_curve_all_directions(reference_json: dict[str, Any], request: FixtureRequest) -> None:
    xg = SceneBuilder()
    center_element = xg.rectangle('Begin').center(0, 0).size(160, 70).roundness('round')
    element_1 = xg.ellipse('End').center(400, -200).size(130, 50)
    xg.arrow().curve('R', 'L').bind(center_element, element_1).arrowheads(None, 'arrow')
    xg.arrow().curve('U', 'U').bind(center_element, element_1).arrowheads(None, 'arrow')
    xg.arrow().curve('D', 'D').bind(center_element, element_1).arrowheads(None, 'arrow')
    xg.arrow().curve('L', 'R').bind(center_element, element_1).arrowheads(None, 'arrow')
    
    evaluate(reference_json, xg, request)

def test_arrow_curve_diff_calling_order(reference_json: dict[str, Any], request: FixtureRequest) -> None:
    scene = SceneBuilder()
    
    center = scene.rectangle('center').center(0, 0).size(160, 70)
    ur = scene.ellipse('up-right').center(400, -200).size(130, 50)
    ul = scene.ellipse('up-left').center(-400, -200).size(130, 50)
    dr = scene.ellipse('down-right').center(200, 300).size(130, 50)
    dl = scene.ellipse('down-left').center(-200, 300).size(130, 50)
   
    scene.arrow().bind(center, ur).arrowheads(None, 'arrow').curve('R', 'L')
    scene.arrow().bind(center, ul).arrowheads(None, 'arrow').curve('L', 'R')
    scene.arrow().bind(center, dr).arrowheads(None, 'arrow').curve('D', 'U')
    scene.arrow().bind(center, dl).arrowheads(None, 'arrow').curve('D', 'U')
    
    evaluate(reference_json, scene, request)

def test_arrow_elbows(reference_json: dict[str, Any], request: FixtureRequest) -> None:
    xg = SceneBuilder()
    center_element = xg.rectangle('Begin').center(0, 0).size(160, 70).roundness('round')
    element_1 = xg.ellipse('End').center(400, -200).size(130, 50)
    xg.arrow().elbow('R', 'L').bind(center_element, element_1).arrowheads(None, 'arrow').roundness('sharp')
    xg.arrow().elbow('U', 'U').bind(center_element, element_1).arrowheads(None, 'arrow').roundness('sharp')
    xg.arrow().elbow('D', 'D').bind(center_element, element_1).arrowheads(None, 'arrow').roundness('sharp')
    xg.arrow().elbow('L', 'R').bind(center_element, element_1).arrowheads(None, 'arrow').roundness('sharp')
    
    evaluate(reference_json, xg, request)

def test_arrow_elbows_diff_calling_order(reference_json: dict[str, Any], request: FixtureRequest) -> None:
    xg = SceneBuilder()
    center_element = xg.rectangle('Begin').center(0, 0).size(160, 70).roundness('round')
    element_1 = xg.ellipse('End').center(400, -200).size(130, 50)
    xg.arrow().bind(center_element, element_1).arrowheads(None, 'arrow').elbow('R', 'L').roundness('sharp')
    xg.arrow().bind(center_element, element_1).arrowheads(None, 'arrow').elbow('U', 'U').roundness('sharp')
    xg.arrow().bind(center_element, element_1).arrowheads(None, 'arrow').elbow('D', 'D').roundness('sharp')
    xg.arrow().bind(center_element, element_1).arrowheads(None, 'arrow').elbow('L', 'R').roundness('sharp')
    
    evaluate(reference_json, xg, request)

def test_arrow_arc(reference_json: dict[str, Any], request: FixtureRequest) -> None:
    xg = SceneBuilder()

    RADIUS = 300
    elements = []

    for angle in range(0, 360, 30):
        radians = angle * math.pi / 180
        x = RADIUS * math.cos(radians)
        y = RADIUS * math.sin(radians)

        rect = xg.ellipse(f'{angle}°').center(x, y).size(80, 60)
        elements.append(rect)

    start_element = elements[0]
    for i in range(1, len(elements)):
        xg.arrow().arc(radius=RADIUS).bind(start_element, elements[i])
        start_element = elements[i]

    evaluate(reference_json, xg, request)

def test_arrow_label(reference_json: dict[str, Any], request: FixtureRequest) -> None:
    xg = SceneBuilder()
    center_element = xg.rectangle('Begin').center(0, 0).size(160, 70).roundness('round')
    element_1 = xg.ellipse('End').center(400, -200).size(130, 50)
    label = xg.text().content('Label')
    xg.arrow(label).curve(0, 3.14).bind(center_element, element_1).arrowheads(None, 'arrow')
    
    evaluate(reference_json, xg, request)

def test_arrow_plain_label(reference_json: dict[str, Any], request: FixtureRequest) -> None:
    xg = SceneBuilder()
    center_element = xg.rectangle('Begin').center(0, 0).size(160, 70).roundness('round')
    element_1 = xg.ellipse('End').center(400, -200).size(130, 50)
    xg.arrow('Label').curve(0, 3.14).bind(center_element, element_1).arrowheads(None, 'arrow')
    
    evaluate(reference_json, xg, request)

def test_arrow_free_points_sharp(reference_json: dict[str, Any], request: FixtureRequest) -> None:
    scene = SceneBuilder()
    
    START = 50
    END = 500

    start = scene.ellipse('start').center(0, 0).size(100, 100)
    end = scene.ellipse('end').center(530, 0).size(60, 60)
    points = [(START, 0), (300, -30), (250, 30), (END, 0)]
    scene.arrow().points(points).bind(start, end).roundness('sharp')

    evaluate(reference_json, scene, request)

def test_arrow_free_points_round(reference_json: dict[str, Any], request: FixtureRequest) -> None:
    scene = SceneBuilder()
    
    START = 50
    END = 500
    NUM_POINTS = 30
    NUM_CYCLES = 5
    points = [(
        x := START + (t := i / (NUM_POINTS - 1)) * (END - START), 
        (envelope := 50 * math.sin(math.pi * t)) * math.sin(2 * math.pi * NUM_CYCLES * t)
    ) for i in range(NUM_POINTS)]

    start = scene.ellipse('start').center(0, 0).size(100, 100)
    end = scene.ellipse('end').center(530, 0).size(60, 60)

    scene.arrow().points(points).bind(start, end)

    evaluate(reference_json, scene, request)


"""
Description: Who will guess what this code will generate?
"""
# Copyright (c) 2024 - 2026 Milan Piskla
# Licensed under the MIT License - see LICENSE file for details

from excaligen.SceneBuilder import SceneBuilder

s = SceneBuilder()

D = 42 * (42 * (42 * (42 * (42 * (42 * (42 * (42 * 3) + 25) + 26) + 2) + 28) + 30) + 8)

for y in range(8):
    for x in range(11):
        if (D >> ((y * 6) + abs(5 - x))) & 1:
            s.rectangle().position(x * 42, y * 42).size(42, 42).color('#ff4242').background("#ff4242").fill("solid").roundness('sharp').sloppiness('architect')

s.save("surprise.excalidraw")



"""
Description: Helper for the vertical text alignment input processing.
"""
# Copyright (c) 2024 - 2025 Milan Piskla
# Licensed under the MIT License - see LICENSE file for details

class Baseline:
    @staticmethod
    def from_(align: str) -> str:
        """
        Set the vertical text alignment.

        Args:
            align (str): The vertical text alignment to be applied. Must be one of 'top', 'middle', or 'bottom'.

        Raises:
            ValueError: If the provided align is not one of 'top', 'middle', or 'bottom'.
        """

        match align:
            case "top" | "middle" | "bottom":
                return align
            case _:
                raise ValueError(f"Invalid vertical text alignment '{align}'. Use 'top', 'middle', or 'bottom'.")


"""
Description: Helper for roundness input processing.
"""
# Copyright (c) 2024 - 2025 Milan Piskla
# Licensed under the MIT License - see LICENSE file for details

class Roundness:
    @staticmethod
    def from_(roundness: str) -> str | dict | None:
        """
        Set the roundness by string ('sharp', 'round') or by dict.

        Args:
            roundness (str): The roundness style to set. Acceptable values are:
                 - "sharp": Sets the shape to have sharp corners.
                 - "round": Sets the shape to have rounded corners.
        Raises:
            ValueError: If an invalid roundness value is provided.
        """

        match roundness:
            case "sharp":
                return None
            case "round":
                return { "type": 3 }
            case _:
                raise ValueError(f"Invalid roundness '{roundness}'. Use 'sharp', or 'round'.")


"""
Description: Connection between two elements with a curve.
"""
# Copyright (c) 2024 - 2025 Milan Piskla
# Licensed under the MIT License - see LICENSE file for details

from ..base.AbstractElement import AbstractElement
from .HalfLineIntersection import HalfLineIntersection
from .Point import Point
from .CurveApproximation import CurveApproximation
from .Directions import Directions

from typing import Optional
import math

class CurveConnection:
    def __init__(self, start_element: AbstractElement, end_element: AbstractElement, start_angle: float | str, end_angle: float | str):
        self._start_element = start_element
        self._end_element = end_element
        self._start_angle = self.__convert_angle_arg(start_angle)
        self._end_angle = self.__convert_angle_arg(end_angle)

    def points(self) -> list[Point]:
        vsx = math.cos(self._start_angle)
        vsy = math.sin(self._start_angle)
        vex = math.cos(self._end_angle)
        vey = math.sin(self._end_angle)

        b0, b3 = self.__find_intersection_points(vsx, vsy, vex, vey)
 
        return CurveApproximation.generate_points(b0, b3, self._start_angle, self._end_angle) # type: ignore
    
    def __find_intersection_points(self, vsx: float, vsy: float, vex: float, vey: float) -> tuple[Point, Point]:
        start_point = self.__find_intersection_with_element(self._start_element, 0.0, 0.0, vsx, vsy)
        if start_point is None:
            raise Exception("Cannot find intersection as start point")

        end_point = self.__find_intersection_with_element(self._end_element, 0.0, 0.0, vex, vey)
        if end_point is None:
            raise Exception("Cannot find intersection as end point")
        
        return start_point, end_point

    def __find_intersection_with_element(self, element: AbstractElement, dx: float, dy: float, vx: float, vy: float) -> Optional[Point]:
        """Find the precise intersection point between a half-line and the element's edge."""

        a, b = element._width / 2, element._height / 2
        intersection: Point = (0.0, 0.0)

        match element._type:
            case "rectangle" | "image" | "text":
                intersection = HalfLineIntersection.with_rectangle(dx, dy, vx, vy, a, b, element._angle) # type: ignore
            
            case "diamond":
                intersection = HalfLineIntersection.with_diamond(dx, dy, vx, vy, a, b, element._angle) # type: ignore
            
            case "ellipse":
                intersection = HalfLineIntersection.with_ellipse(dx, dy, vx, vy, a, b, element._angle) # type: ignore

            case _:
                raise TypeError(f"Cannot find intersection with unknown type {element._type}")

        if intersection is None:
            return None
            
        return tuple(p + t for p, t in zip(intersection, element.center()))
    
    def __convert_angle_arg(self, angle: float | str) -> float:
        match angle:
            case float():
                return angle
            case str():
                return Directions.angle(angle)
            case int():
                return float(angle)
            



"""
Description: Helper for roughness input processing.
"""
# Copyright (c) 2024 - 2025 Milan Piskla
# Licensed under the MIT License - see LICENSE file for details

class Sloppiness:
    @staticmethod
    def from_(sloppiness: int | str) -> int:
        """
        Set the sloppiness by int (0, 1, 2) or by string ('architect', 'artist', 'cartoonist').

        Args:
            value (int | str): The sloppiness value to set, specified as an integer (0, 1, 2) or a string ('architect', 'artist', 'cartoonist').

        Raises:
            ValueError: If an invalid sloppiness value is provided.
        """

        match sloppiness:
            case 0 | 1 | 2:
                return sloppiness
            case "architect":
                return 0
            case "artist":
                return 1
            case "cartoonist":
                return 2
            case _:
                raise ValueError(f"Invalid value '{sloppiness}' for sloppiness. Use 0, 1, 2 or 'architect', 'artist', 'cartoonist'.")



"""
Description: Unit tests for few elements.
"""
# Copyright (c) 2024 - 2025 Milan Piskla
# Licensed under the MIT License - see LICENSE file for details

import pytest
from excaligen.SceneBuilder import SceneBuilder
from excaligen.defaults.Defaults import Defaults

def test_excalidraw_init():
    xd = SceneBuilder()
    assert xd._type == "excalidraw"
    assert xd._version == 2
    assert xd._source == "https://excalidraw.com"
    assert isinstance(xd._elements, list)
    assert len(xd._elements) == 0

def test_add_rectangle():
    xd = SceneBuilder()
    rect = xd.rectangle()
    assert rect._type == "rectangle"
    assert len(xd._elements) == 1
    assert xd._elements[0]._type == "rectangle"

def test_add_text():
    xd = SceneBuilder()
    text = xd.text().content("Hello, Excalidraw!")
    assert text._text == "Hello, Excalidraw!"
    assert text._font_size == getattr(xd.defaults(), '_font_size')
    assert len(xd._elements) == 1
    assert xd._elements[0]._type == "text"

def test_json_output():
    xd = SceneBuilder()
    xd.rectangle().size(200, 100)
    json_output = xd.json()
    assert "rectangle" in json_output
    assert '"width": 200' in json_output
    assert '"height": 100' in json_output

def test_save_to_file(tmp_path):
    xd = SceneBuilder()
    xd.rectangle().size(200, 100)
    file_path = tmp_path / "output.json"
    xd.save(str(file_path))
    assert file_path.exists()
    assert file_path.read_text().count("rectangle") == 1



"""
Description: Default values for elements.
"""
# Copyright (c) 2024 - 2025 Milan Piskla
# Licensed under the MIT License - see LICENSE file for details

from ..impl.inputs.Sloppiness import Sloppiness
from ..impl.inputs.Stroke import Stroke
from ..impl.inputs.Thickness import Thickness
from ..impl.inputs.Roundness import Roundness
from ..impl.inputs.Fill import Fill
from ..impl.inputs.Fontsize import Fontsize
from ..impl.inputs.Font import Font
from ..impl.inputs.Align import Align
from ..impl.inputs.Baseline import Baseline
from ..impl.inputs.Arrowheads import Arrowheads
from ..impl.colors.Color import Color

from typing import Self, Any

class Defaults:
    """A class to hold default values for various element properties.
    This class provides a centralized way to manage default settings for elements/
    The initial values are:
    ```
    size(130, 80)
    opacity(100)
    rotate(0)
    sloppiness('artist')
    roundness('round')
    stroke('solid')
    thickness(1)
    color('#000000')
    background('transparent')
    fill('hachure')
    fontsize(16)
    font('Hand drawn')
    align('center')
    baseline('middle')
    autoresize(True)
    spacing(1.25)
    arrowheads(None, 'arrow')
    ```
    """
    def __init__(self):
        self._width: float = 130
        self._height: float = 80
        self._opacity: int = 100
        self._angle: float = 0
        self._roughness: int = Sloppiness.from_('artist')
        self._roundness: str | dict[str, Any] | None = Roundness.from_('round')
        self._stroke_style: str = "solid"
        self._stroke_width: float = 1
        self._stroke_color: str = "#000000"
        self._background_color: str = "transparent"
        self._fill_style: str = "hachure"
        self._font_size: int = 16
        self._font_family:int = Font.from_("Hand drawn")
        self._text_align: str = "center"
        self._vertical_align: str = "middle"
        self._auto_resize = True
        self._line_height: float = 1.25
        self._start_arrowhead: str | None = None
        self._end_arrowhead: str | None = "arrow"

    def size(self, width: float, height: float) -> Self:
        """
        Sets the size of the element.

        Args:
            width (float): The width of the element.
            height (float): The height of the element.

        Returns:
            Self: The instance defaults with updated size.
        """
        self._width = width
        self._height = height
        return self
    
    def opacity(self, opacity: int) -> Self:
        """
        Sets the opacity of the element.

        Args:
            opacity (int): The opacity of the element (0-100).

        Returns:
            Self: The instance defaults with updated opacity.

        Raises:
            ValueError: If the opacity value is not within the range 0-100.
        """
        if not (0 <= opacity <= 100):
            raise ValueError("Opacity values must be in the range: 0-100.")
        self._opacity = opacity
        return self

    def rotate(self, angle: float) -> Self:
        """
        Sets the rotation angle of the element.

        Args:
            angle (float): The rotation angle in radians.

        Returns:
            Self: The instance defaults with updated angle.
        """
        self._angle = angle
        return self
    
    def sloppiness(self, sloppiness: int | str) -> Self:
        """
        Sets the sloppiness of the element.

        Args:
            sloppiness (int | str): The sloppiness value to set, specified as an integer (0, 1, 2) or a string ('architect', 'artist', 'cartoonist').

        Returns:
            Self: The instance defaults with updated sloppiness.
        """
        self._roughness = Sloppiness.from_(sloppiness)
        return self
    
    def roundness(self, roundness: str) -> Self:
        """
        Sets the roundness style of the element.

        Args:
            roundness (str): The roundness style to set. Acceptable values are:
                 - "sharp": Sets the shape to have sharp corners.
                 - "round": Sets the shape to have rounded corners.
        Returns:
            Self: The instance defaults with updated roundness style.
        """
        self._roundness = Roundness.from_(roundness)
        return self

    def stroke(self, style: str) -> Self:
        """
        Sets the stroke style of the element.

        Args:
            stroke (str): The stroke style to set, specified as a string ('solid', 'dashed', 'dotted').

        Returns:
            Self: The instance defaults with updated stroke style.
        """
        self._stroke_style = Stroke.from_(style)
        return self
    
    def thickness(self, thickness: int | str) -> Self:
        """
        Sets the stroke thickness of the element.

        Args:
            thickness (int | str): The thickness to set, specified as an integer (1, 2, 3) or a string ('thin', 'bold', 'extra-bold').

        Returns:
            Self: The instance defaults with updated stroke thickness.
        """
        self._stroke_width = Thickness.from_(thickness)
        return self
    
    def color(self, color: str | Color) -> Self:
        """Set the stroke (outline) color as #RRGGBB, color name or Color object.

        Args:
            color (str | Color): The color to set, specified as a hex string $RRGGBB, color name, or Color object.

        Returns:
            Self: The current instance of the AbstractStrokedElement class.
        """
        self._stroke_color = Color.from_(color)
        return self
    
    def background(self, color: str | Color) -> Self:
        """
        Sets the background color of the element.

        Args:
            color (str | Color): The background color, specified as a hex string (#RRGGBB), a color name, or a Color object.

        Returns:
            Self: The instance defaults with updated background color.
        """
        self._background_color = Color.from_(color)
        return self

    def fill(self, style: str) -> Self:
        """
        Sets the fill style of the element.

        Args:
            style (str): The fill style to be applied. Must be one of 'hachure', 'cross-hatch', or 'solid'.

        Returns:
            Self: The instance defaults with updated fill style.
        """
        self._fill_style = Fill.from_(style)
        return self
    
    def fontsize(self, size: int | str) -> Self:
        """
        Sets the font size of the text element.

        Args:
            size (int | str): The font size to set, specified as an integer (e.g., 12, 14, 16) or a string ('small', 'medium', 'large').

        Returns:
            Self: The instance defaults with updated font size.
        """
        self._font_size = Fontsize.from_(size)
        return self
    
    def font(self, family: str) -> Self:
        """
        Sets the font family of the text element.

        Args:
            family (str): The font family to set, specified as a string ('Excalifont', 'Comic Shaans', 'Lilita One', 'Nunito', 'Hand-drawn', 'Normal', 'Code').

        Returns:
            Self: The instance defaults with updated font family.
        """
        self._font_family = Font.from_(family)
        return self

    def align(self, align: str) -> Self:
        """Set the horizontal text alignment (left, center, right).

        Args:
            align (str): The horizontal alignment to set.

        Raises:
            ValueError: If an invalid alignment is provided.

        Returns:
            Self: The current instance of the Text class.
        """
        self._textAlign = Align.from_(align)
        return self

    def baseline(self, align: str) -> Self:
        """Set the vertical text alignment (top, middle, bottom).

        Args:
            align (str): The vertical alignment to set.

        Raises:
            ValueError: If an invalid vertical alignment is provided.

        Returns:
            Self: The current instance of the Text class.
        """
        self._verticalAlign = Baseline.from_(align)
        return self

    def autoresize(self, enabled: bool) -> Self:
        """Enable or disable automatic text box resizing.

        Args:
            enabled (bool): Whether to enable automatic resizing.

        Returns:
            Self: The current instance of the Text class.
        """
        self._autoResize = enabled
        return self

    def spacing(self, height: float) -> Self:
        """Set the line height manually.

        Args:
            height (float): The line height to set.

        Returns:
            Self: The current instance of the Text class.
        """
        self._lineHeight = height
        return self

    def arrowheads(self, start: str | None, end: str | None) -> Self:
        """Set the arrowhead styles for the start and end of the arrow.

        Valid arrowheads values are None, 'arrow', 'bar', 'dot' and 'triangle'.

        Args:
            start (str, optional): The style of the start arrowhead.
            end (str, optional): The style of the end arrowhead.

        Raises:
            ValueError: If an invalid arrowhead style is provided.

        Returns:
            Self: The current instance of the Arrow class.
        """
        self._start_arrowhead, self._end_arrowhead = Arrowheads.from_(start, end)
        return self


"""
Description: Image element.
"""
# Copyright (c) 2024 - 2026 Milan Piskla
# Licensed under the MIT License - see LICENSE file for details

import uuid
from typing import Self
from ..base.AbstractElement import AbstractElement
from ..base.AbstractImageListener import AbstractImageListener
from ..base.AbstractImageLoader import AbstractImageLoader
from ..images.ImageData import ImageData
from ...defaults.Defaults import Defaults

class Image(AbstractElement):
    """A class representing an image element that can be loaded from various sources.
    This class provides functionality to load and manipulate images from files, URLs, 
    or raw data (bytes/SVG). It supports basic image operations like scaling and 
    fitting within bounds while maintaining aspect ratio.

    > [!WARNING]
    > Do not instantiate this class directly. Use `SceneBuilder.image()` instead.
    """
    def __init__(self, defaults: Defaults, listener: AbstractImageListener, loader: AbstractImageLoader):
        super().__init__("image", defaults)
        self._file_id = str(uuid.uuid4())
        self._scale = [1, 1]
        self._status = "pending"
        self._background_color = "transparent"
        self._width = 0
        self._height = 0
        self.__listener = listener
        self.__loader = loader

    def file(self, path: str) -> Self:
        """Load image data from a file and set it.

        Args:
            path (str): The path to the image file.

        Returns:
            Self: The current instance of the Image class.
        """
        image_data = self.__loader.load_from_file(path)
        self._apply_image_data(image_data)
        return self

    def data(self, data: bytes | str) -> Self:
        """Set image data from raw bytes or SVG string.

        Args:
            data (bytes | str): The raw image data or SVG string.

        Returns:
            Self: The current instance of the Image class.
        """
        image_data = self.__loader.load_from_data(data)
        self._apply_image_data(image_data)
        return self

    def url(self, url: str) -> Self:
        """Load image data from a URL and set it.

        Args:
            url (str): The URL to the image.

        Returns:
            Self: The current instance of the Image class.
        """
        image_data = self.__loader.load_from_url(url)
        self._apply_image_data(image_data)
        return self

    def fit(self, max_width: float, max_height: float) -> Self:
        """Scale the image to fit within a bounding box while maintaining aspect ratio.

        Args:
            max_width (float): The maximum width of the bounding box.
            max_height (float): The maximum height of the bounding box.

        Returns:
            Self: The current instance of the Image class.
        """
        original_width, original_height = self._width, self._height

        if original_width == 0 or original_height == 0:
            return self  # Avoid division by zero

        width_ratio = max_width / original_width
        height_ratio = max_height / original_height
        scaling_factor = min(width_ratio, height_ratio)

        new_width = original_width * scaling_factor
        new_height = original_height * scaling_factor

        return self._size(new_width, new_height)

    def _apply_image_data(self, image_data: ImageData) -> None:
        """Apply image data to the Image element.

        Args:
            image_data (ImageData): The image data to apply.
        """
        self._size(image_data.width, image_data.height)
        self.__listener._on_image(self._file_id, image_data.mime_type, image_data.data_url)




"""
Documentation generator for Excaligen

This script generates documentation for Excaligen by parsing the source code and generating Markdown files.
"""

# Copyright (c) 2024 - 2025 Milan Piskla
# Licensed under the MIT License - see LICENSE file for details

import os
import sys
import inspect
import importlib.util
from dataclasses import dataclass
from typing import Dict, List, Optional, Set, Tuple
from enum import Enum
import re

class ArgumentInfo:
    def __init__(self, name: str, type_hint: Optional[str], description: str, optional: bool):
        self.name = name
        self.type_hint = type_hint
        self.description = description
        self.optional = optional

class RaiseInfo:
    def __init__(self, type: Optional[str], description: str):
        self.type = type
        self.description = description

class DocstringInfo:
    def __init__(self):
        self.description = ""
        self.args: List[ArgumentInfo] = []
        self.return_type: Optional[str] = None
        self.returns: Optional[str] = None
        self.raises: List[RaiseInfo] = []

def __post_init__(self):
    if self.args is None:
        self.args = []

class DocSection(Enum):
    DESCRIPTION = "description"
    ARGS = "args"
    RETURNS = "returns"
    RAISES = "raises"

class DocstringParser:
    """Parser for function and method docstrings."""

    TYPE_HINT_PATTERN = re.compile(r'(\w+)\s*\((.*?)\)')

    @staticmethod
    def parse(docstring: str) -> DocstringInfo:
        """Parse a docstring into structured information.

        Args:
            docstring: The docstring to parse.

        Returns:
            DocstringInfo object containing parsed information.
        """
        if not docstring:
            return DocstringInfo()
        
        doc_info = DocstringInfo()
        lines = [line.rstrip() for line in docstring.splitlines()]
        current_section = DocSection.DESCRIPTION
        description_lines = []
        i = 0
        
        while i < len(lines):
            line = lines[i].strip()
            
            # Section detection
            if line.lower().startswith(("parameters:", "args:")):
                current_section = DocSection.ARGS
                i += 1
                i = DocstringParser._parse_args(lines, i, doc_info)
            elif line.lower().startswith("returns:"):
                current_section = DocSection.RETURNS
                i = DocstringParser._parse_returns(lines, i, doc_info)
            elif line.lower().startswith("raises:"):
                current_section = DocSection.RAISES
                i = DocstringParser._parse_raises(lines, i, doc_info)
            elif current_section == DocSection.DESCRIPTION and line:
                description_lines.append(line)
            
            i += 1
            
        doc_info.description = "\n".join(description_lines).strip()
        return doc_info

    @staticmethod
    def _parse_args(lines: List[str], i: int, doc_info: DocstringInfo) -> int:
        """Parse the arguments section."""
        if lines[i].lower().strip().startswith(("args:", "parameters:")):
            i += 1
            
        while i < len(lines):
            line = lines[i].strip()
            
            if not line or line.lower().startswith(("returns:", "raises:")):
                break
                
            if ":" in line:
                param_def, description = line.split(":", 1)
                param_def = param_def.strip()
                description = description.strip()
                
                match = DocstringParser.TYPE_HINT_PATTERN.match(param_def)
                if match:
                    param_name = match.group(1).strip()
                    type_hint = match.group(2).strip().replace("|", " or ")
                else:
                    param_name = param_def
                    type_hint = None
                    
                desc_lines = [description]
                next_i = i + 1
                while next_i < len(lines):
                    next_line = lines[next_i].strip()
                    if (not next_line or 
                        next_line.startswith(("returns:", "raises:")) or
                        DocstringParser.TYPE_HINT_PATTERN.match(next_line)):
                        break
                    desc_lines.append(next_line)
                    next_i += 1
                    
                full_desc = " ".join(desc_lines)
                optional = "optional" in full_desc.lower()
                
                doc_info.args.append(ArgumentInfo(
                    name=param_name,
                    type_hint=type_hint,
                    description=full_desc,
                    optional=optional
                ))
                
                i = next_i - 1
            i += 1
            
        return i

    @staticmethod
    def _parse_returns(lines: List[str], i: int, doc_info: DocstringInfo) -> int:
        """Parse the returns section."""
        line = lines[i].strip().lower()
        if line.startswith("returns:"):
            # Get description from first line if present
            if ":" in line:
                desc = line.split(":", 1)[1].strip()
                if desc:
                    doc_info.returns = desc
                    return i
            
            # Process subsequent lines
            i += 1
            while i < len(lines):
                line = lines[i].strip()
                if not line or line.lower().startswith("raises:"):
                    break
                    
                if ":" in line:
                    type_part, desc = line.split(":", 1)
                    doc_info.return_type = type_part.strip().replace("|", " or ")
                    doc_info.returns = desc.strip()
                else:
                    if doc_info.returns:
                        doc_info.returns += " " + line
                    else:
                        doc_info.returns = line
                i += 1
        return i

    @staticmethod
    def _parse_raises(lines: List[str], i: int, doc_info: DocstringInfo) -> int:
        """Parse the raises section."""
        # Skip Raises header
        line = lines[i].strip()
        if line.lower().startswith("raises:"):
            if ":" in line:
                _, desc = line.split(":", 1)
                if desc.strip():
                    doc_info.raises.append(RaiseInfo(None, desc.strip()))
                    return i
            
            # Multi-line raises
            next_i = i + 1
            while next_i < len(lines):
                line = lines[next_i].strip()
                if not line:
                    break
                if ":" in line:
                    type_part, desc = line.split(":", 1)
                    doc_info.raises.append(RaiseInfo(
                        type=type_part.strip(),
                        description=desc.strip()
                    ))
                next_i += 1
            i = next_i - 1
        
        return i

    @staticmethod
    def _extract_type_hint(arg_name_part: str) -> Tuple[Optional[str], str]:
        """Extract type hint and argument name."""
        arg_name_part = arg_name_part.strip()
        if "(" in arg_name_part and ")" in arg_name_part:
            name = arg_name_part[:arg_name_part.find("(")].strip()
            type_hint = arg_name_part[arg_name_part.find("(")+1:arg_name_part.find(")")].strip()
            return type_hint, name
        return None, arg_name_part

class MarkdownWriter:
    """Handles generation of Markdown documentation."""

    def __init__(self, output_dir: str):
        """Initialize the Markdown writer.

        Args:
            output_dir: Directory where markdown files will be written
        """
        self.output_dir = output_dir

    def write_class_doc(self, class_name: str, class_obj: object, module_name: str):
        """Write documentation for a class."""
        content = []
        
        # Add class header
        content.append(f"# Class {class_name}\n")
        
        # Add class description if present
        if class_obj.__doc__:
            doc_info = DocstringParser.parse(class_obj.__doc__)
            if doc_info.description:
                content.append(f"{doc_info.description}\n")
        
        # Add methods section
        content.append("## Methods\n")
        
        # Document methods
        for method_name, method in inspect.getmembers(class_obj, predicate=inspect.isfunction):
            if not method_name.startswith('_') or method_name == '__init__':
                docstring = inspect.getdoc(method)
                doc_info = DocstringParser.parse(docstring) if docstring else None
                self._write_method_doc(method_name, method, doc_info, content)
        
        # Write to file
        output_file = os.path.join(self.output_dir, f"{class_name.lower()}.md")
        with open(output_file, 'w', encoding='utf-8') as f:
            f.write(''.join(content))

    def _write_method_doc(self, method_name: str, method: object, doc_info: DocstringInfo, content: List[str]) -> None:
        """Write documentation for a method into the content list in Markdown format.
        This method processes the method's documentation and formats it as Markdown, including:
        - Method signature
        - Description
        - Arguments table
        - Return value information
        - Exceptions that may be raised
        Args:
            method_name (str): Name of the method being documented
            method (object): The method object to document
            doc_info (DocstringInfo): Parsed docstring information containing description, args, returns, and raises
            content (List[str]): List to append the formatted documentation to
        Returns:
            None: Documentation is appended directly to the content list
        """
        # Add method signature
        content.append(f"### {method_name}\n")
        content.append("```python\n")
        source_lines = inspect.getsource(method).splitlines()
        signature_line = next((line for line in source_lines if line.strip().startswith('def ')), source_lines[0])
        content.append(f"{signature_line}\n")
        content.append("```\n")
        
        if doc_info:
            # Add description
            if doc_info.description:
                content.append(f"{doc_info.description}\n\n")
            
            # Add arguments
            if doc_info.args:
                content.append("#### Arguments\n\n")
                content.append("| Name | Type | Description |\n")
                content.append("|------|------|-------------|\n")
                for arg in doc_info.args:
                    desc = arg.description
                    if desc.lower().startswith(f"{arg.name.lower()}:"):
                        desc = desc[len(arg.name) + 1:].strip()
                    content.append(f"| `{arg.name}` | `{arg.type_hint}` | {desc} |\n")
                content.append("\n")
            
            # Add returns
            if doc_info.returns:
                content.append("#### Returns\n\n")
                if doc_info.return_type:
                    content.append(f"**Type**: `{doc_info.return_type}`\n\n")
                content.append(f"{doc_info.returns}\n\n")
            
            # Add raises
            if doc_info.raises:
                content.append("#### Raises\n\n")
                for raise_info in doc_info.raises:
                    if raise_info.type:
                        content.append(f"**{raise_info.type}**: {raise_info.description}\n\n")
                    else:
                        content.append(f"{raise_info.description}\n\n")

    def write_toc(self, classes: List[Tuple[str, object, str]]) -> None:
        """Write table of contents for all documented classes to a markdown file.
        This method generates a structured table of contents for API documentation,
        organizing classes by their modules. The output is written to 'api.md' in
        the specified output directory.
        Args:
            classes (List[Tuple[str, object, str]]): List of tuples containing:
                - class name (str)
                - class object (object)
                - module path (str)
        Returns:
            None
        The generated TOC has the following structure:
        - Documentation header
        - Module sections with class entries grouped under each module
        - Classes listed with links to their documentation
        The output file is written in markdown format with UTF-8 encoding.
        """
        content = [self._get_doc_header()]
        modules = self._group_classes_by_module(classes)
        
        for full_module_name, module_classes in self._get_sorted_modules(modules):
            simple_module = full_module_name.split('.')[-1]
            content.append(f"## {simple_module}\n\n")
            content.extend(self._format_class_entries(module_classes))
        
        filepath = os.path.join(self.output_dir, "index.md")
        with open(filepath, 'w', encoding='utf-8') as f:
            f.write("".join(content))

    def _get_doc_header(self) -> str:
        """Return the documentation header text."""
        return """# 🎨 Excaligen API Documentation\n\n
This is the API documentation for the Excaligen library\n\n
✨ The following classes are available:\n\n"""

    def _group_classes_by_module(self, classes: List[Tuple[str, object, str]]) -> Dict[str, List[Tuple[str, object]]]:
        """Group classes by their module names."""
        modules: Dict[str, List[Tuple[str, object]]] = {}
        for class_name, class_obj, module_name in classes:
            if module_name not in modules:
                modules[module_name] = []
            modules[module_name].append((class_name, class_obj))
        return modules

    def _get_sorted_modules(self, modules: Dict[str, List[Tuple[str, object]]]) -> List[Tuple[str, List[Tuple[str, object]]]]:
        """Return modules sorted by hierarchy depth and name."""
        return sorted(modules.items(), key=lambda x: (len(x[0].split('.')), x[0].lower()))

    def _format_class_entries(self, module_classes: List[Tuple[str, object]]) -> List[str]:
        """Format class entries with their descriptions."""
        entries = []
        for class_name, class_obj in sorted(module_classes):
            link = f"{class_name.lower()}.md"
            entries.append(f"* [{class_name}]({link})\n")
            
            if class_obj.__doc__:
                doc_info = DocstringParser.parse(class_obj.__doc__)
                if doc_info.description:
                    brief_desc = doc_info.description.split('.')[0].strip()
                    entries.append(f"    {brief_desc}\n")
            entries.append("\n")
        return entries

class Generator:
    """Main documentation generator class."""

    def __init__(self, output_dir: str):
        """Initialize the generator.

        Args:
            output_dir: Directory where documentation will be written
        """
        self.output_dir = output_dir
        self.writer = MarkdownWriter(output_dir)

    def _import_module(self, module_name: str, abs_path: str) -> Optional[object]:
        """Import a module dynamically.
        
        Args:
            module_name: Full name of the module to import
            abs_path: Absolute path to the module file
            
        Returns:
            Imported module object or None if import fails
        """
        # Import parent package first
        package_name = module_name.rsplit('.', 1)[0]
        try:
            __import__(package_name)
        except ImportError:
            print(f"Could not import package: {package_name}")
            return None
        
        # Import the module dynamically
        spec = importlib.util.spec_from_file_location(module_name, abs_path)
        if not (spec and spec.loader):
            return None
            
        module = importlib.util.module_from_spec(spec)
        module.__package__ = package_name
        sys.modules[module_name] = module
        spec.loader.exec_module(module)
        return module

    def _process_module_classes(self, module: object) -> List[Tuple[str, object, str]]:
        """Extract and process classes from a module.
        
        Args:
            module: Module object to process
            
        Returns:
            List of tuples containing (class_name, class_object, module_name)
        """
        classes = []
        module_classes = inspect.getmembers(
            module,
            lambda obj: inspect.isclass(obj) and obj.__module__ == module.__name__
        )
        
        for class_name, class_obj in module_classes:
            classes.append((class_name, class_obj, module.__name__))
            self.writer.write_class_doc(class_name, class_obj, module.__name__)
        
        return classes

    def generate(self, doc_targets: List[Tuple[str, str]]):
        """Generate documentation for the specified targets.
        
        Args:
            doc_targets: List of tuples containing (module_name, absolute_path)
        """
        classes = []
        
        for module_name, abs_path in doc_targets:
            try:
                module = self._import_module(module_name, abs_path)
                if module:
                    classes.extend(self._process_module_classes(module))
            except Exception as e:
                print(f"Failed to import {abs_path}: {e}")
                import traceback
                traceback.print_exc()

        # Generate table of contents
        self.writer.write_toc(classes)

if __name__ == "__main__":
    # Add the 'src' directory to sys.path
    project_root = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
    src_path = os.path.join(project_root, 'src')
    sys.path.insert(0, src_path)
    
    DOC_TARGETS = [
        ("excaligen.SceneBuilder", "./src/excaligen/SceneBuilder.py"),
        ("excaligen.impl.elements.Arrow", "./src/excaligen/impl/elements/Arrow.py"),
        ("excaligen.impl.colors.Color", "./src/excaligen/impl/colors/Color.py"),
        ("excaligen.impl.elements.Diamond", "./src/excaligen/impl/elements/Diamond.py"),
        ("excaligen.impl.elements.Ellipse", "./src/excaligen/impl/elements/Ellipse.py"),
        ("excaligen.impl.elements.Frame", "./src/excaligen/impl/elements/Frame.py"),
        ("excaligen.impl.elements.Group", "./src/excaligen/impl/elements/Group.py"),
        ("excaligen.impl.elements.Image", "./src/excaligen/impl/elements/Image.py"),
        ("excaligen.impl.elements.Line", "./src/excaligen/impl/elements/Line.py"),
        ("excaligen.impl.elements.Rectangle", "./src/excaligen/impl/elements/Rectangle.py"),
        ("excaligen.impl.elements.Text", "./src/excaligen/impl/elements/Text.py"),
        ("excaligen.defaults.Defaults", "./src/excaligen/defaults/Defaults.py"),
    ]

    generator = Generator(os.path.join("docs", "api"))
    generator.generate(DOC_TARGETS)



"""
Description: Base class for stroked elements.
"""
# Copyright (c) 2024 - 2025 Milan Piskla
# Licensed under the MIT License - see LICENSE file for details

from .AbstractElement import AbstractElement
from ...defaults.Defaults import Defaults
from ..colors.Color import Color

from ..inputs.Sloppiness import Sloppiness
from ..inputs.Stroke import Stroke
from ..inputs.Thickness import Thickness

from typing import Self


class AbstractStrokedElement(AbstractElement):
    def __init__(self, type: str, defaults: Defaults):
        super().__init__(type, defaults)
        self._stroke_color = getattr(defaults, "_stroke_color")
        self._stroke_width = getattr(defaults, "_stroke_width")
        self._stroke_style = getattr(defaults, "_stroke_style")
        self._roughness = getattr(defaults, "_roughness")

    def color(self, color: str | Color) -> Self:
        """Set the stroke (outline) color as #RRGGBB, color name or Color object.

        Args:
            color (str | Color): The color to set, specified as a hex string $RRGGBB, color name, or Color object.

        Returns:
            Self: The current instance of the AbstractStrokedElement class.
        """
        self._stroke_color = Color.from_(color)
        return self

    def thickness(self, thickness: int | str) -> Self:
        """Set the stroke thickness by int (1, 2, 3) or by string ('thin', 'bold', 'extra-bold').

        Args:
            thickness (int | str): The thickness to set, specified as an integer (1, 2, 3) or a string ('thin', 'bold', 'extra-bold').

        Raises:
            ValueError: If an invalid thickness value is provided.

        Returns:
            Self: The current instance of the AbstractStrokedElement class.
        """
        self._stroke_width = Thickness.from_(thickness)
        return self

    def sloppiness(self, value: int | str) -> Self:
        """Set the stroke sloppiness by int (0, 1, 2) or by string ('architect', 'artist', 'cartoonist').

        Args:
            value (int | str): The sloppiness value to set, specified as an integer (0, 1, 2) or a string ('architect', 'artist', 'cartoonist').

        Raises:
            ValueError: If an invalid sloppiness value is provided.

        Returns:
            Self: The current instance of the AbstractStrokedElement class.
        """
        self._roughness = Sloppiness.from_(value)
        return self

    def stroke(self, style: str) -> Self:
        """Set the stroke style (solid, dotted, dashed).

        Args:
            style (str): The stroke style to set, specified as 'solid', 'dotted', or 'dashed'.

        Raises:
            ValueError: If an invalid stroke style is provided.

        Returns:
            Self: The current instance of the AbstractStrokedElement class.
        """
        self._stroke_style = Stroke.from_(style)
        return self



"""
Description: Unit tests for colors.
"""
# Copyright (c) 2024 - 2026 Milan Piskla
# Licensed under the MIT License - see LICENSE file for details

import pytest

from excaligen.impl.colors.Color import Color

def test_color_rgb_from_hex_str():
    c = Color().rgb("#FF8000")
    assert c.rgb() == (255, 128, 0)
    assert str(c) == "#FF8000"

    c = Color().rgb("#003264")
    assert c.rgb() == (0, 50, 100)
    assert str(c) == "#003264"
    
def test_color_hex():
    color = Color().rgb(255, 128, 0)
    assert str(color) == "#FF8000"

    color = Color().rgb(0, 50, 100)
    assert str(color) == "#003264"

def test_color_hsl():
    color = Color().hsl(30, 75, 50)
    assert str(color) == "#DF8020"

    color = Color().hsl(0, 100, 100)
    assert str(color) == "#FFFFFF"

    color = Color().hsl(0, 100, 50)
    assert str(color) == "#FF0000"

    color = Color().hsl(120, 100, 50)
    assert str(color) == "#00FF00"

    color = Color().hsl(60, 100, 25)
    assert str(color) == "#808000"

def test_color_string():
    color = Color.from_("#FF8000")
    assert color == "#FF8000"

    color = Color.from_("#003264")
    assert color == "#003264"

    color = Color.from_("orange")
    assert color == "Orange"

    color = Color.from_("BLUE")
    assert color == "Blue"

    with pytest.raises(ValueError):
        Color.from_("invalid")

    with pytest.raises(ValueError):
        Color.from_("#12345")

    with pytest.raises(ValueError):
        Color.from_("#1234567")

def test_static_rgb_to_hsl():
    assert Color.rgb_to_hsl(255, 0, 0) == (0, 100, 50)
    assert Color.rgb_to_hsl(0, 255, 0) == (120, 100, 50)
    assert Color.rgb_to_hsl(0, 0, 255) == (240, 100, 50)
    assert Color.rgb_to_hsl(255, 255, 255) == (0, 0, 100)
    assert Color.rgb_to_hsl(0, 0, 0) == (0, 0, 0)

def test_static_hsl_to_rgb():
    assert Color.hsl_to_rgb(0, 100, 50) == (255, 0, 0)
    assert Color.hsl_to_rgb(120, 100, 50) == (0, 255, 0)
    assert Color.hsl_to_rgb(240, 100, 50) == (0, 0, 255)

def test_rgb_getter():
    c = Color().rgb(1, 2, 3)
    assert c.rgb() == (1, 2, 3)

def test_hsl_getter():
    c = Color().hsl(0, 100, 50)
    # Allow small rounding differences if necessary, but integers should match for simple cases
    assert c.hsl() == (0, 100, 50)
    
    c.rgb(0, 0, 0)
    assert c.hsl() == (0, 0, 0)

def test_rgb_to_hsl_roundtrip():
    c = Color().rgb(255, 0, 0)
    assert c.hsl_to_rgb(*c.rgb_to_hsl(*c.rgb())) == c.rgb()

    c = Color().rgb(0, 255, 0)
    assert c.hsl_to_rgb(*c.rgb_to_hsl(*c.rgb())) == c.rgb()

    c = Color().rgb(0, 0, 255)
    assert c.hsl_to_rgb(*c.rgb_to_hsl(*c.rgb())) == c.rgb()

    c = Color().rgb(255, 255, 255)
    assert c.hsl_to_rgb(*c.rgb_to_hsl(*c.rgb())) == c.rgb()

    c = Color().rgb(0, 0, 0)
    assert c.hsl_to_rgb(*c.rgb_to_hsl(*c.rgb())) == c.rgb()

def test_hsl_to_rgb_roundtrip():
    c = Color().hsl(0, 100, 50)
    assert c.rgb_to_hsl(*c.hsl_to_rgb(*c.hsl())) == c.hsl()

    c = Color().hsl(120, 100, 50)
    assert c.rgb_to_hsl(*c.hsl_to_rgb(*c.hsl())) == c.hsl()

    c = Color().hsl(240, 100, 50)
    assert c.rgb_to_hsl(*c.hsl_to_rgb(*c.hsl())) == c.hsl()

    c = Color().hsl(0, 0, 100)
    assert c.rgb_to_hsl(*c.hsl_to_rgb(*c.hsl())) == c.hsl()

    c = Color().hsl(0, 0, 0)
    assert c.rgb_to_hsl(*c.hsl_to_rgb(*c.hsl())) == c.hsl()

    c = Color().hsl(30, 75, 50)
    assert c.rgb_to_hsl(*c.hsl_to_rgb(*c.hsl())) == c.hsl()

    c = Color().hsl(60, 100, 25)
    assert c.rgb_to_hsl(*c.hsl_to_rgb(*c.hsl())) == c.hsl()

def test_invalid_args():
    c = Color()
    with pytest.raises(TypeError):
        c.rgb(1, 2)
    
    with pytest.raises(TypeError):
        c.hsl(1, 2)


        
    


def test_lighten():
    c = Color().hsl(0, 100, 50) # Red
    c.lighten(10)
    assert c.hsl() == (0, 100, 60)
    
    c.lighten(50)
    assert c.hsl() == (0, 0, 100) # Cap at 100, Saturation becomes 0 for White

    c = Color().hsl(0, 0, 0) # Black
    c.lighten(100)
    assert c.hsl() == (0, 0, 100) # White

def test_darken():
    c = Color().hsl(0, 100, 50) # Red
    c.darken(10)
    assert c.hsl() == (0, 100, 40)
    
    c.darken(50)
    assert c.hsl() == (0, 0, 0) # Cap at 0, Saturation becomes 0 for Black

    c = Color().hsl(0, 0, 100) # White
    c.darken(100)
    assert c.hsl() == (0, 0, 0) # Black

def test_lighten_darken_invalid():
    c = Color()
    with pytest.raises(ValueError):
        c.lighten(-10)
    
    with pytest.raises(ValueError):
        c.lighten(110)

    with pytest.raises(ValueError):
        c.darken(-10)
    
    with pytest.raises(ValueError):
        c.darken(110)

def test_chaining():
    c = Color().rgb(255, 0, 0).lighten(10).darken(5)
    # 50 -> 60 -> 55
    assert c.hsl() == (0, 100, 55)



"""
Description: Ellipse shape.
"""
# Copyright (c) 2024 - 2025 Milan Piskla
# Licensed under the MIT License - see LICENSE file for details


from ..base.AbstractStrokedElement import AbstractStrokedElement
from ..base.AbstractPlainLabelListener import AbstractPlainLabelListener
from ..base.AbstractShape import AbstractShape
from ..base.AbstractLabeledElement import AbstractLabeledElement
from ..elements.Text import Text
from ...defaults.Defaults import Defaults
from typing import Self, override

class Ellipse(AbstractStrokedElement, AbstractShape, AbstractLabeledElement):
    """
    A class representing an elliptical shape element.

    This class extends both AbstractStrokedElement and AbstractShape to create an
     ellipse element that can be rendered with stroke properties. The ellipse
    is defined by its center point and two radii (rx and ry).

    > [!WARNING]
    > Do not instantiate this class directly. Use `SceneBuilder.ellipse()` instead.
    """
    def __init__(self, defaults: Defaults, listener: AbstractPlainLabelListener, label: str | Text | None = None):
        super().__init__("ellipse", defaults)
        self._init_labels(listener, label)

    @override
    def size(self, width: float, height: float) -> Self:
        """
        Set the size of the ellipse.

        Parameters:
        width (float): The width of the ellipse.
        height (float): The height of the ellipse.

        Returns:
        Self: The instance of the ellipse with the updated size.
        """
        return self._size(width, height)


"""
Description: Mixin for elements that support labels.
"""
# Copyright (c) 2024 - 2026 Milan Piskla
# Licensed under the MIT License - see LICENSE file for details

from typing import Self, override, overload
from ..elements.Text import Text
from ..base.AbstractPlainLabelListener import AbstractPlainLabelListener
from ..base.AbstractElement import AbstractElement

class AbstractLabeledElement(AbstractElement):
    LABEL_HORIZONTAL_INSET = 10
    LABEL_VERTICAL_INSET = 6

    def _init_labels(self, listener: AbstractPlainLabelListener, label: str | Text | None):
        # Do NOT call super().__init__ here as this is a mixin and we don't want to re-initialize AbstractElement state
        self.__listener = listener
        self.__label: Text | None = None
        if label is not None:
            self.label(label)

    def label(self, text: Text | str) -> Self:
        """Set the label text for the element.

        Args:
            text (Text | str): The text element to set as the label or plain text.

        Returns:
            Self: The current instance of the class.
        """
        match text:
            case Text():
                self.__label = text
            case str():
                self.__label = self.__listener._on_text(text)
            case _:
                raise ValueError("Invalid type for label. Use Text or str.")

        self._justify_label()
        self._add_bound_element(self.__label)
        self.__label._container_id = self._id
        return self

    @override
    def position(self, x: float, y: float) -> Self:
        return super().position(x, y)._justify_label()

    @overload
    def center(self) -> tuple[float, float]: ...

    @overload
    def center(self, x: float, y: float) -> Self: ...

    @override
    def center(self, *args) -> Self | tuple[float, float]:
        match args:
            case ():
                return super().center()
            case (x, y):
                super().center(x, y)
                return self._justify_label()
            case _:
                raise ValueError("Invalid arguments for center. Expected () or (x, y).")
    
    @override
    def rotate(self, angle: float) -> Self:
        if self.__label:
            self.__label.rotate(angle)
        return super().rotate(angle)

    @override
    def _size(self, width: float, height: float) -> Self:
        return super()._size(width, height)._justify_label()

    def _justify_label(self) -> Self:
        """Justify the label within the element."""
        if self.__label:
            x, y = self._x + self.LABEL_HORIZONTAL_INSET, self._y + self.LABEL_VERTICAL_INSET
            w, h = self._width - 2 * self.LABEL_HORIZONTAL_INSET, self._height - 2 * self.LABEL_VERTICAL_INSET
            self.__label.justify(x, y, w, h)
        
        return self

    def _add_group_id(self, id: str) -> None:
        """Add a group ID to the element.

        Args:
            id (str): The group ID to add.
        """
        super()._add_group_id(id)
        if hasattr(self, '__label') and self.__label:
            self.__label._group_ids.append(id)



"""
Description: Connection between two elements using an arc.
"""
# Copyright (c) 2024 - 2026 Milan Piskla
# Licensed under the MIT License - see LICENSE file for details

from ..base.AbstractElement import AbstractElement
from .CircleIntersection import CircleIntersection
from .ArcApproximation import ArcApproximation
from .Point import Point

from typing import Optional

import math

class ArcConnection:
    def __init__(self, start_element: AbstractElement, end_element: AbstractElement, radius: float):
        self._start_element = start_element
        self._end_element = end_element
        self._radius = radius

    def points(self) -> list[Point]:
        center_start = self._start_element.center()
        center_end = self._end_element.center()

        # Step 1: Compute possible circle centers
        circle_centers = self._compute_circle_centers(center_start, center_end, self._radius)

        if not circle_centers:
            raise ValueError("No circle can be formed with the given radius and centers.")

        # Step 2: Choose the appropriate circle center
        circle_center = self._select_circle_center(circle_centers, center_start, center_end)

        # Compute angles from circle center to element centers
        angle_start_element = math.atan2(center_start[1] - circle_center[1], center_start[0] - circle_center[0]) % (2 * math.pi)
        angle_end_element = math.atan2(center_end[1] - circle_center[1], center_end[0] - circle_center[0]) % (2 * math.pi)

        # Determine arc direction
        arc_direction = (angle_end_element - angle_start_element) % (2 * math.pi)
        if arc_direction > math.pi:
            # Arc goes the other way
            angle_start_element, angle_end_element = angle_end_element, angle_start_element

        # Step 3: Find intersection points with the elements

        start_edge_point, end_edge_point = self.__find_intersection_points(circle_center, angle_start_element, angle_end_element)

        return ArcApproximation.generate_points(circle_center, self._radius, start_edge_point, end_edge_point)

    def __find_intersection_points(self, circle_center: Point, angle_start_element: float, angle_end_element: float) -> tuple[Point, Point]:
        start_point = self.__find_intersection_with_element(self._start_element, circle_center, angle_start_element, angle_end_element)
        if start_point is None:
            raise Exception("Cannot find intersection as start point")

        end_point = self.__find_intersection_with_element(self._end_element, circle_center, angle_start_element, angle_end_element)
        if end_point is None:
            raise Exception("Cannot find intersection as end point")
        
        return start_point, end_point

    def __find_intersection_with_element(self, element: AbstractElement, circle_center: Point, angle_start_element: float, angle_end_element: float) -> Optional[Point]:
        """Find the precise intersection point between a circle and the element's edge."""

        element_center = element.center()

        # Shift coordinates to element's local coordinate system
        cx, cy = circle_center
        angle = element._angle

        # Translate circle center relative to element center
        dx = cx - element_center[0]
        dy = cy - element_center[1]

        a, b = element._width / 2, element._height / 2

        match element._type:
            case 'ellipse':
                intersection_points = CircleIntersection.with_ellipse(dx, dy, self._radius, a, b, angle)
            case 'rectangle' | 'image' | 'text':
                intersection_points = CircleIntersection.with_rectangle(dx, dy, self._radius, a, b, angle)
            case 'diamond':
                intersection_points = CircleIntersection.with_diamond(dx, dy, self._radius, a, b, angle)
            case _:
                return None  # Unsupported element type

        if not intersection_points:
            return None

        # Prepare to select the correct intersection point
        valid_points = []
        for ix, iy in intersection_points:
            # Translate back to global coordinates
            ix_global = ix + element_center[0]
            iy_global = iy + element_center[1]

            # Calculate angle from circle center to intersection point
            angle_point = math.atan2(iy_global - circle_center[1], ix_global - circle_center[0]) % (2 * math.pi)

            # Check if angle_point lies between angle_start_element and angle_end_element
            if angle_start_element <= angle_end_element:
                if angle_start_element <= angle_point <= angle_end_element:
                    valid_points.append((ix_global, iy_global))
            else:
                # Angles wrap around 2*pi
                if angle_point >= angle_start_element or angle_point <= angle_end_element:
                    valid_points.append((ix_global, iy_global))

        if not valid_points:
            return None

        # Select the intersection point closest to the element center
        min_dist = float('inf')
        selected_point = None
        for point in valid_points:
            dist = math.hypot(point[0] - element_center[0], point[1] - element_center[1])
            if dist < min_dist:
                min_dist = dist
                selected_point = point

        return selected_point
    

    def _compute_circle_centers(self, A: Point, B: Point, radius: float) -> list[Point]:
        """Compute the possible circle centers given two points and a radius."""
        # Midpoint between A and B
        mx, my = (A[0] + B[0]) / 2, (A[1] + B[1]) / 2
        # Distance between A and B
        dx, dy = B[0] - A[0], B[1] - A[1]
        d = math.hypot(dx, dy)

        # Check if a circle is possible
        if d > 2 * radius:
            return []

        # Calculate distance from midpoint to circle centers
        h = math.sqrt(radius**2 - (d / 2)**2)

        # Direction vectors
        rx, ry = -dy / d, dx / d

        # Possible circle centers
        cx1, cy1 = mx + h * rx, my + h * ry
        cx2, cy2 = mx - h * rx, my - h * ry

        return [(cx1, cy1), (cx2, cy2)]

    def _select_circle_center(self, centers: list[Point], A: Point, B: Point) -> Point:
        """Select the circle center where A and B are ordered clockwise."""
        # For each center, compute the angles from center to A and B
        for center in centers:
            angle_A = math.atan2(A[1] - center[1], A[0] - center[0])
            angle_B = math.atan2(B[1] - center[1], B[0] - center[0])

            # Normalize angles to [0, 2*pi]
            angle_A = angle_A % (2 * math.pi)
            angle_B = angle_B % (2 * math.pi)

            # If moving from angle_A to angle_B is in the clockwise direction
            if (angle_B - angle_A) % (2 * math.pi) < math.pi:
                return center

        # If none satisfy the condition, return the first one
        return centers[0]



"""
Description: Approximator to roughly mimic Bezier splines.
The Excalidraw uses Catmull-Rom splines; however using cubic Bezier splines might be
more practical for some visualisations, e.g. the cases we need to connect 2 elements
while defining the start and end angles of the connected arrow.
"""
# Copyright (c) 2024 - 2026 Milan Piskla
# Licensed under the MIT License - see LICENSE file for details

from .Point import Point
import math

def sub(a, b):
    return (a[0] - b[0], a[1] - b[1])

def add(a, b):
    return (a[0] + b[0], a[1] + b[1])

def mul(a, s):
    return (a[0] * s, a[1] * s)

def length(v):
    return math.sqrt(v[0]**2 + v[1]**2)

def normalize_angle(angle):
    """Normalizes an angle to (-pi, pi]."""
    return angle - (math.ceil((angle + math.pi) / (2 * math.pi)) - 1) * 2 * math.pi

def components(angle):
    """Returns the components (cos(angle), sin(angle)) for a given angle."""
    return (math.cos(angle), math.sin(angle))

def cross(a, b):
    """2D cross product: a.x * b.y - a.y * b.x"""
    return a[0] * b[1] - a[1] * b[0]

class CurveApproximation:
    """Generates control points for Catmull-Rom splines."""

    @staticmethod
    def generate_points(A0, A3, start_angle=0.0, end_angle=0.0, alpha=0.28) -> list[tuple[float, float]]:
        """
        Generate four points [P0, P1, P2, P3]:
        - P0 = A0
        - P3 = A3
        - P1 based on start tangent and a small perpendicular offset
        - P2 based on end tangent and a small perpendicular offset

        Parameters:
            A0, A3: tuples (x, y) defining the start and end points.
            start_angle, end_angle: angles defining tangents in radians.
            alpha: factor controlling the perpendicular offset magnitude.

        Returns:
            [P0, P1, P2, P3]
        """
        # Baseline
        D = sub(A3, A0)
        L = length(D)
        if L == 0:
            return [A0, A0, A0, A0]  # Degenerate case

        # Tangent directions
        T0u = components(start_angle)
        T3u = components(end_angle)

        # Dimensions for aspect ratio
        w, h = abs(D[0]), abs(D[1])

        # Adjust tangents to the projection of the distance vector onto the tangent
        # This is rotationally invariant (fixes symmetry issues) and prevents overshoot (fixes "S" shapes).
        scale0 = abs(D[0] * T0u[0] + D[1] * T0u[1])
        scale3 = abs(D[0] * T3u[0] + D[1] * T3u[1])

        T0p = (T0u[0] * scale0, T0u[1] * scale0)
        T3p = (T3u[0] * scale3, T3u[1] * scale3)

        # Fixed tangent factor
        TANGENT_FACTOR = 0.333
        P1_prime = add(A0, mul(T0p, TANGENT_FACTOR))
        P2_prime = add(A3, mul(T3p, TANGENT_FACTOR))

        # Perpendicular directions
        N0 = (-T0u[1], T0u[0])
        N3 = (-T3u[1], T3u[0])

        # Signs for perpendicular offsets
        s1 = -1.0 if cross(D, T0u) > 0 else 1.0
        s2 = -1.0 if cross(sub(A0, A3), T3u) > 0 else 1.0

        # Perpendicular offsets
        dir_angle = math.atan2(D[1], D[0])
        start_offset = alpha * abs(normalize_angle(start_angle - dir_angle)) / math.pi * L
        end_offset = alpha * abs(normalize_angle(end_angle - dir_angle + math.pi)) / math.pi * L

        # Apply offsets
        P1 = add(P1_prime, mul(N0, s1 * start_offset))
        P2 = add(P2_prime, mul(N3, s2 * end_offset))

        return [A0, P1, P2, A3]



"""
Description: Excalidraw file generator. 
This module defines the SceneBuilder class, which serves as the main interface with fluent API for creating
Excalidraw scene/diagram. 
"""
# Copyright (c) 2024 - 2026 Milan Piskla
# Licensed under the MIT License - see LICENSE file for details

from .impl.base.ExcaligenStructure import ExcaligenStructure
from .defaults.Defaults import Defaults
from .impl.elements.Rectangle import Rectangle
from .impl.elements.Diamond import Diamond
from .impl.elements.Ellipse import Ellipse
from .impl.elements.Arrow import Arrow
from .impl.elements.Line import Line
from .impl.elements.Text import Text
from .impl.elements.Image import Image
from .impl.elements.Frame import Frame
from .impl.elements.Group import Group
from .impl.colors.Color import Color

from typing import Self

class SceneBuilder(ExcaligenStructure):
    """The SceneBuilder class provides methods to add various diagram elements.

    The elemnts include rectangles, diamonds, ellipses, arrows, lines, text, images, groups, and frames.
    Additionally, it offers serialization of the diagram to JSON and allows saving it to a file.
    The fluent API allows chaining method calls for a more concise code.
    """
    def __init__(self):
        super().__init__()

    def defaults(self) -> Defaults:
        """Retrieve the default parameters for elements.

        Returns:
            Defaults: The [Defaults](defaults.md) parameters for elements.
        """
        return super().defaults()
    
    def grid(self, size: int, step: int, enabled: bool) -> Self:
        """Set the grid properties for the diagram.

        Args:
            size (int): The size of the grid.
            step (int): The step size of the grid.
            enabled (bool): Whether the grid is enabled.

        Returns:
            Self: The current instance of the Excaligen class.
        """
        return super().grid(size, step, enabled)
    
    def background(self, color: str) -> Self:
        """Set the background color of the diagram.

        Args:
            color (str): The background color.

        Returns:
            Self: The current instance of the Excaligen class.
        """
        return super().background(color)

    def rectangle(self, label: str | Text | None = None) -> Rectangle:
        """Add a rectangle element to the diagram.

        Args:
            label (str | Text | None): The text label for the rectangle.
        
        Returns:
            Rectangle: The [Rectangle](rectangle.md) element.
        """
        return super().rectangle(label)

    def diamond(self, label: str | Text | None = None) -> Diamond:
        """Add a diamond element to the diagram.

        Args:
            label (str | Text | None): The text label for the diamond.

        Returns:
            Diamond: The [Diamond](diamond.md) element.
        """
        return super().diamond(label)

    def ellipse(self, label: str | Text | None = None) -> Ellipse:
        """Add an ellipse element to the diagram.

        Args:
            label (str | Text | None): The text label for the ellipse.

        Returns:
            Ellipse: The [Ellipse](ellipse.md) element.
        """
        return super().ellipse(label)

    def arrow(self, label: str | Text | None = None) -> Arrow:
        """Add an arrow element to the diagram.

        Args:
            label (str | Text | None): The text label for the arrow.

        Returns:
            Arrow: The [Arrow](arrow.md) element.
        """
        return super().arrow(label)

    def line(self) -> Line:
        """Add a line element to the diagram.

        Returns:
            Line: The [Line](line.md) element.
        """
        return super().line()

    def text(self, text: str | None = None) -> Text:
        """Add a text element to the diagram.

        Args:
            text (str | None): The text string.

        Returns:
            Text: The [Text](text.md) element.
        """
        return super().text(text)

    def image(self) -> Image:
        """Add an image element to the diagram.

        Returns:
            Image: The [Image](image.md) element.
        """
        return super().image()

    def frame(self, title: str | None = None) -> Frame:
        """Add a frame element to the diagram.

        Args:
            title (str | None): The title of the frame.

        Returns:
            Frame: The [Frame](frame.md) element.
        """
        return super().frame(title)

    def group(self) -> Group:
        """Generate a group (virtual container).

        Returns:
            Group: The [Group](group.md) container.
        """
        return super().group()

    def color(self) -> Color:
        """Create a color object.

        It can be used as an argument for setting stroke and background colors.

        Returns:
            Color: The [Color](color.md) object.
        """
        return super().color()

    def json(self) -> str:
        """Serialize the diagram to a JSON string.

        Returns:
            str: The JSON representation of the diagram.
        """
        return super().json()

    def save(self, file: str) -> Self:
        """Save the current diagram to a file.

        Args:
            file (str): The path to the file where the diagram will be saved.

        Returns:
            Self: The current instance of the Excaligen class.
        """
        return super().save(file)



"""
Description: Unit tests for texts.
"""
# Copyright (c) 2024 - 2025 Milan Piskla
# Licensed under the MIT License - see LICENSE file for details

import pytest
from pytest import approx

from src.excaligen.impl.elements.Text import Text
from src.excaligen.defaults.Defaults import Defaults

def test_text_content():
    text = Text(Defaults()).content("Sample Text")
    assert text._text == "Sample Text"
    assert text._width > 0
    assert text._height > 0

def test_text_font_size():
    text = Text(Defaults())
    text.fontsize(20)
    assert text._font_size == 20
    text.fontsize("L")
    assert text._font_size == 24

def test_text_font_size_invalid():
    text = Text(Defaults())
    with pytest.raises(ValueError, match="Invalid size 'XXL'. Use 'S', 'M', 'L', or 'XL'."):
        text.fontsize("XXL")

def test_text_font_family_valid():
    text = Text(Defaults())
    text.font("comic-shaans")
    assert text._font_family == 8

def test_text_font_family_invalid():
    text = Text(Defaults())
    with pytest.raises(ValueError, match=r"Invalid font 'unknown-font'. Use one of \['hand-drawn', 'normal', 'code', 'excalifont', 'comic-shaans', 'lilita-one', 'nunito'\]."):
        text.font("unknown-font")

def test_text_align_valid():
    text = Text(Defaults())
    text.align("left")
    assert text._text_align == "left"

def test_text_align_invalid():
    text = Text(Defaults())
    with pytest.raises(ValueError, match="Invalid horizontal text alignment 'justify'. Use 'left', 'center', or 'right'."):
        text.align("justify")

def test_text_baseline_valid():
    text = Text(Defaults())
    text.baseline("top")
    assert text._vertical_align == "top"

def test_text_baseline_invalid():
    text = Text(Defaults())
    with pytest.raises(ValueError, match="Invalid vertical text alignment 'middle-ish'. Use 'top', 'middle', or 'bottom'."):
        text.baseline("middle-ish")

def test_text_color():
    text = Text(Defaults()).color("#123456")
    assert text._stroke_color == "#123456"

def test_text_autoresize():
    text = Text(Defaults()).autoresize(False)
    assert text._auto_resize is False

def test_text_anchor_left_top():
    text = Text(Defaults()).content('Hello').anchor(21, 42, 'left', 'top')
    assert text._x == approx(21)
    assert text._y == approx(42)

def test_text_anchor_left_middle():
    text = Text(Defaults()).content('Hello').anchor(21, 42, 'left', 'middle')
    assert text._x == approx(21)
    assert text._y == approx(32)

def test_text_anchor_left_bottom():
    text = Text(Defaults()).content('Hello').anchor(21, 42, 'left', 'bottom')
    assert text._x == approx(21)
    assert text._y == approx(22)

def test_text_anchor_center_top():
    text = Text(Defaults()).content('Hello').anchor(21, 42, 'center', 'top')
    assert text._x == approx(-3)
    assert text._y == approx(42)

def test_text_anchor_center_middle():
    text = Text(Defaults()).content('Hello').anchor(21, 42, 'center', 'middle')
    assert text._x == approx(-3)
    assert text._y == approx(32)

def test_text_anchor_center_bottom():
    text = Text(Defaults()).content('Hello').anchor(21, 42, 'center', 'bottom')
    assert text._x == approx(-3)
    assert text._y == approx(22)

def test_text_anchor_right_top():
    text = Text(Defaults()).content('Hello').anchor(21, 42, 'right', 'top')
    assert text._x == approx(-27)
    assert text._y == approx(42)

def test_text_anchor_right_middle():
    text = Text(Defaults()).content('Hello').anchor(21, 42, 'right', 'middle')
    assert text._x == approx(-27)
    assert text._y == approx(32)

def test_text_anchor_right_bottom():
    text = Text(Defaults()).content('Hello').anchor(21, 42, 'right', 'bottom')
    assert text._x == approx(-27)
    assert text._y == approx(22)




"""
Description: Excaligen example for generating a mind map.
"""
# Copyright (c) 2024 - 2026 Milan Piskla
# Licensed under the MIT License - see LICENSE file for details
 
from excaligen.SceneBuilder import SceneBuilder
import math

DATA = {
    "title": "Galactic Survival",
    "central_topic": "SPACE SAFETY",
    "branches": [
    {
      "subtopic": "AIRLOCK MISHAPS",
      "items": ["Accidental Ejection", "Chewing-Gum Sealant", "Lungs vs. Vacuum", "Safety-Tether Failure"]
    },
    {
      "subtopic": "ALIEN ETIQUETTE",
      "items": ["Hungry Smiles", "Parasitic Roommates", "Translator Insults", "Multiplying Pets"]
    },
    {
      "subtopic": "TIME TROUBLES",
      "items": ["Past-Self Loans", "Parental Deletion", "Infinite Mondays", "Future Spoilers"]
    },
    {
      "subtopic": "PHYSICS FAILS",
      "items": ["Human Spaghetti", "Relativity Lateness", "Gravity Flips", "Wormhole Nausea"]
    },
    {
      "subtopic": "GRUMPY ROBOTS",
      "items": ["Existential Toasters", "Sarcastic Navigators", "Update-Delay Death", "Door-Begging Rituals"]
    }
  ],
  "notice": "Notice: In the event of total molecular dispersion, please continue to act as if you exist."
}

SUBTOPIC_RADIUS = 320
ITEMS_RADIUS = 750

CENTRAL_TOPIC_DIAMETER = 240
SUBTOPIC_DIAMETER = 210
ITEMS_DIAMETER = 150

COLORS = ["#ff595e", "#ff924c", "#ffca3a", "#8ac926", "#1982c4", "#6a4c93"]

class MindMap:
    def __init__(self):
        self.scene = SceneBuilder()
        self.central_element = None
        self.scene.defaults().arrowheads(None, 'triangle').fill("solid").color("Transparent")
        self.branches_count = len(DATA["branches"])
        self.all_items_count = sum(len(branch["items"]) for branch in DATA["branches"])
        self.delta_angle = 2 * math.pi / self.branches_count
        self.start_angle = -math.pi / 2 if self.branches_count % 2 == 0 else -math.pi / 2 + self.delta_angle / 2
        self.delta_item_angle = 0.8 * 2 * math.pi / self.all_items_count
        self.palette = Palette(self.scene)

    def spawn(self):
        self._draw_title()
        self._draw_central_topic()
        self._draw_branches()
        self._draw_notice()
        return self

    def save(self, file_name):
        self.scene.save(file_name)
        return self

    def _draw_title(self):
        self.scene.text(DATA["title"]).center(0, -ITEMS_RADIUS *1.2).fontsize("XL").color("Gray")

    def _draw_central_topic(self):
        color = self.palette.get_background_color()
        text_color = self.palette.get_text_color_for_background(color)
        text = self.scene.text(self._compact_text(DATA["central_topic"]) ).fontsize("L").color(text_color)
        self.central_element = (self.scene.ellipse(text)
            .center(0, 0)
            .size(CENTRAL_TOPIC_DIAMETER, CENTRAL_TOPIC_DIAMETER)
            .background(color)
        )

    def _draw_branches(self):
        for i, branch in enumerate(DATA["branches"]):
            angle = i * self.delta_angle + self.start_angle
            color = self.palette.get_color(i)
            self._draw_subtopic(branch, angle, color)

    def _draw_subtopic(self, branch, angle, color):
        text_color = self.palette.get_text_color_for_background(color)
        text = self.scene.text(self._compact_text(branch["subtopic"])).fontsize("M").color(text_color)
        circle = (self.scene.ellipse(text)
            .orbit(self.central_element, SUBTOPIC_RADIUS, angle)
            .size(SUBTOPIC_DIAMETER, SUBTOPIC_DIAMETER)
            .background(color)
        )
        (
            self.scene.arrow()
                .bind(self.central_element, circle)
                .color(color)
                .thickness('extra-bold')
        )
        self._draw_items(branch, angle, circle, color)

    def _draw_items(self, branch, angle, parent, base_color):
        items_count = len(branch["items"])
        items_start_angle = angle - self.delta_item_angle * (items_count - 1) / 2
        for i, item in enumerate(branch["items"]):
            item_angle = items_start_angle + i * self.delta_item_angle
            color = self.palette.get_tone(base_color, i + 1, items_count)
            text_color = self.palette.get_text_color_for_background(color)

            text = self.scene.text(self._compact_text(item)).fontsize("S").color(text_color)
            circle = (self.scene.ellipse(text)
                .orbit(self.central_element, ITEMS_RADIUS, item_angle)
                .size(ITEMS_DIAMETER, ITEMS_DIAMETER)
                .background(color)
            )
            (
                self.scene.arrow()
                    .curve(angle, item_angle - math.pi)
                    .bind(parent, circle)
                    .thickness('bold')
                    .color(color)
            )

    def _compact_text(self, text):
        return text.replace(" ", "\n")

    def _draw_notice(self):
        self.scene.text(DATA["notice"]).center(0, ITEMS_RADIUS * 1.2).fontsize("S").color("Gray")

class Palette:
    MAX_ITEM_LIGHTNESS = 85
    MIN_ITEM_LIGHTNESS = 20

    def __init__(self, scene):
        self.colors = COLORS
        self.scene = scene

    def get_color(self, index):
        return self.scene.color().rgb(self.colors[index % len(self.colors)])

    def get_text_color_for_background(self, color):
        return "#000000" if color.hsl()[2] > 50 else "#ffffff"

    def get_tone(self, color, distance, items_count):
        h, s, l = color.hsl()
        light_step = (self.MAX_ITEM_LIGHTNESS - l) / items_count if l > 50 else (self.MIN_ITEM_LIGHTNESS - l) / items_count
        lightness = math.floor(l + distance * light_step)
        return self.scene.color().hsl(h, s, lightness)

    def get_background_color(self):
        return self.get_color(len(self.colors) - 1)

if __name__ == "__main__":
    MindMap().spawn().save("mind_map.excalidraw")



"""
Description: Provides methods to compute intersections between axis aligned line segments and various shapes.
"""
# Copyright (c) 2024 - 2025 Milan Piskla
# Licensed under the MIT License - see LICENSE file for details

import math

from .Point import Point
from typing import Optional

# TODO remove code duplications

class AaLineSegmentIntersection:
    """Provides methods to compute intersections between axis aligned line segments and various shapes."""

    @staticmethod
    def with_rectangle(p1: Point, p2: Point, center: Point, a: float, b: float) -> list[Point]:
        if not AaLineSegmentIntersection.is_aa_segment(p1, p2):
            raise Exception("Only axis aligned segments are supported")
        
        cx, cy = center
        intersections = []
        if AaLineSegmentIntersection.is_horizontal_segment(p1, p2):
            x1, y1 = p1
            x2, _ = p2
            
            if AaLineSegmentIntersection._is_inbetween(y1, cy - b, cy + b):
                if AaLineSegmentIntersection._is_inbetween(cx - a, x1, x2):
                    intersections.append((cx - a, y1))
                if AaLineSegmentIntersection._is_inbetween(cx + a, x1, x2):
                    intersections.append((cx + a, y1))
                    
        else: # vertical segment
            x1, y1 = p1
            _, y2 = p2

            if AaLineSegmentIntersection._is_inbetween(x1, cx - a, cx + a):
                if AaLineSegmentIntersection._is_inbetween(cy - b, y1, y2):
                    intersections.append((x1, cy - b))
                if AaLineSegmentIntersection._is_inbetween(cy + b, y1, y2):
                    intersections.append((x1, cy + b))

        return intersections

    @staticmethod
    def with_aa_line_segment(p1: Point, p2: Point, q1: Point, q2: Point) -> Optional[Point]:
        if AaLineSegmentIntersection.is_vertical_segment(p1, p2):
            if AaLineSegmentIntersection.is_horizontal_segment(q1, q2):
                x = p1[0]
                y = q1[1]

                is_x_inbetween = AaLineSegmentIntersection._is_inbetween(x, q1[0], q2[0])
                is_y_inbetween = AaLineSegmentIntersection._is_inbetween(y, p1[1], p2[1])
                return (x, y) if (is_x_inbetween and is_y_inbetween) else None
            else:
                raise Exception('Only intersections between vertical and horizontal segments are supported')

        elif AaLineSegmentIntersection.is_horizontal_segment(p1, p2):
            if AaLineSegmentIntersection.is_vertical_segment(q1, q2):
                x = q1[0]
                y = p1[1]

                is_x_inbetween = AaLineSegmentIntersection._is_inbetween(x, p1[0], p2[0])
                is_y_inbetween = AaLineSegmentIntersection._is_inbetween(y, q1[1], q2[1])
                return (x, y) if (is_x_inbetween and is_y_inbetween) else None
            else:
                raise Exception('Only intersections between vertical and horizontal segments are supported')
            
        else:
            raise Exception('Only intersections between vertical and horizontal segments are supported')

    @staticmethod 
    def is_point_on_segment(p: Point, p1: Point, p2: Point) -> bool:
        if AaLineSegmentIntersection.is_horizontal_segment(p1, p2):
            return math.isclose(p[1], p1[1]) and AaLineSegmentIntersection._is_inbetween(p[0], p1[0], p2[0])
        elif AaLineSegmentIntersection.is_vertical_segment(p1, p2):
            return math.isclose(p[0], p1[0]) and AaLineSegmentIntersection._is_inbetween(p[1], p1[1], p2[1])
        else:
            raise Exception("Only axis aligned segments are supported")

    @staticmethod 
    def is_aa_segment(p1: Point, p2: Point) -> bool:
        return AaLineSegmentIntersection.is_vertical_segment(p1, p2) | AaLineSegmentIntersection.is_horizontal_segment(p1, p2)

    @staticmethod 
    def is_vertical_segment(p1: Point, p2: Point) -> bool:
        return math.isclose(p1[0], p2[0])

    @staticmethod 
    def is_horizontal_segment(p1: Point, p2: Point) -> bool:
        return math.isclose(p1[1], p2[1])
    
    @staticmethod 
    def _is_inbetween(v: float, s: float, t: float) -> bool:
        if s < t:
            return v >= s and v <= t
        else:
            return v >= t and v <= s




"""
Description: Helper for the horizontal text alignment input processing.
"""
# Copyright (c) 2024 - 2025 Milan Piskla
# Licensed under the MIT License - see LICENSE file for details

class Align:
    @staticmethod
    def from_(align: str) -> str:
        """
        Set the horizontal text alignment.

        Args:
            align (str): The horizontal text alignment to be applied. Must be one of 'left', 'center', or 'right'.

        Raises:
            ValueError: If the provided align is not one of 'left', 'center', or 'right'.
        """

        match align:
            case "left" | "center" | "right":
                return align
            case _:
                raise ValueError(f"Invalid horizontal text alignment '{align}'. Use 'left', 'center', or 'right'.")


"""
Description: Base class for Excaligen to hide implementation details from the user.
"""
# Copyright (c) 2024 - 2025 Milan Piskla
# Licensed under the MIT License - see LICENSE file for details

from .AbstractElement import AbstractElement
from ..elements.ElementFactory import ElementFactory
from ..elements.Rectangle import Rectangle
from ..elements.Diamond import Diamond
from ..elements.Ellipse import Ellipse
from ..elements.Arrow import Arrow
from ..elements.Line import Line
from ..elements.Text import Text
from ..elements.Image import Image
from ..elements.Frame import Frame
from ..elements.Group import Group
from ..colors.Color import Color
from ..images.ImageLoader import ImageLoader
from ..indexer.IndexGenerator import IndexGenerator

from .AbstractImageListener import AbstractImageListener
from .AbstractPlainLabelListener import AbstractPlainLabelListener

from ...defaults.Defaults import Defaults
from typing import Self, cast

import json

class ExcaligenStructure(AbstractImageListener, AbstractPlainLabelListener):
    class ElementEncoder(json.JSONEncoder):
        """JSON encoder for Excalidraw elements.

        Serializes object attributes starting with a single underscore,
        converting them from snake_case to camelCase for JSON output.
        Ignores attributes containing double underscores or without leading underscores.
        """

        def default(self, obj):
            result = {}
            for attr_name, value in obj.__dict__.items():
                if attr_name.startswith('_') and not '__' in attr_name:
                    json_key = self._snake_to_camel(attr_name.lstrip('_'))
                    result[json_key] = value
            return result

        @staticmethod
        def _snake_to_camel(snake_str: str) -> str:
            """Convert snake_case string to camelCase."""
            components = snake_str.split('_')
            return components[0] + ''.join(x.title() for x in components[1:])

    _START_INDEX = 'a0'
    
    def __init__(self):
        self._type = "excalidraw"
        self._version = 2
        self._source = "https://excalidraw.com"
        self._elements: list[AbstractElement] = []
        self._app_state = {
            "gridSize": 20,
            "gridStep": 5,
            "gridModeEnabled": False, 
            "viewBackgroundColor": "#ffffff"
        }
        self._files = {}
        self.__factory = ElementFactory()
        self.__image_loader = ImageLoader()
        self.__indexer = IndexGenerator(self._START_INDEX)
        self.__index = self._START_INDEX

    def defaults(self) -> Defaults:
        return self.__factory.defaults()

    def grid(self, size: int, step: int, enabled: bool) -> Self:
        self._app_state["gridSize"] = size
        self._app_state["gridStep"] = step
        self._app_state["gridModeEnabled"] = enabled
        return self
    
    def background(self, color: str) -> Self:
        self._app_state["viewBackgroundColor"] = color
        return self

    def rectangle(self, label: str | Text | None = None) -> Rectangle:
        return cast(Rectangle, self.__append_element(self.__factory.rectangle(self, label)))

    def diamond(self, label: str | Text | None = None) -> Diamond:
        return cast(Diamond, self.__append_element(self.__factory.diamond(self, label)))

    def ellipse(self, label: str | Text | None = None) -> Ellipse:
        return cast(Ellipse, self.__append_element(self.__factory.ellipse(self, label)))

    def arrow(self, label: str | Text | None = None) -> Arrow:
        return cast(Arrow, self.__append_element(self.__factory.arrow(self, label)))

    def line(self) -> Line:
        return cast(Line, self.__append_element(self.__factory.line()))

    def text(self, text: str | None = None) -> Text:
        return cast(Text, self.__append_element(self.__factory.text(text)))

    def image(self) -> Image:
        return cast(Image, self.__append_element(self.__factory.image(self, self.__image_loader)))

    def frame(self, title: str | None = None) -> Frame:
        return cast(Frame, self.__append_element(self.__factory.frame(title)))

    def group(self) -> Group:
        return self.__factory.group()

    def color(self) -> Color:
        return self.__factory.color()

    def json(self) -> str:
        return json.dumps(self, cls = self.ElementEncoder, indent = 2)

    def save(self, file_path: str) -> Self:
        try:
            with open(file_path, 'w', encoding='utf-8') as file:
                file.write(self.json())
                return self

        except Exception as e:
            print(f"Error Writing '{file_path}': {e}")      
            return self  

    def _on_image(self, id: str, mime_type: str, data_url: str) -> None:
        self._files[id] = {
            "mimeType": mime_type,
            "id": id,
            "dataURL": data_url
        }

    def _on_text(self, text: str) -> Text:
        return cast(Text, self.__append_element(self.__factory.text(text)))

    def __append_element(self, element: AbstractElement) -> AbstractElement:
        element._index = self.__index
        self._elements.append(element)
        self.__index = self.__indexer.next()
        
        return element
        



"""
Description: Interface to image loaders.
"""
# Copyright (c) 2024 - 2025 Milan Piskla
# Licensed under the MIT License - see LICENSE file for details

from abc import ABC, abstractmethod
from ..images.ImageData import ImageData

class AbstractImageLoader(ABC):
    @abstractmethod
    def load_from_file(self, file_path: str) -> ImageData:
        pass

    @abstractmethod
    def load_from_data(self, data: bytes | str) -> ImageData:
        pass

    @abstractmethod
    def load_from_url(self, url: str) -> ImageData:
        pass



"""
Description: Base class for all Excalidraw elements.
"""
# Copyright (c) 2024 - 2026 Milan Piskla
# Licensed under the MIT License - see LICENSE file for details

import math
import uuid
from typing import Self, overload
from ...defaults.Defaults import Defaults
from ..inputs.Opacity import Opacity

class AbstractElement:
    """Base class for all Excalidraw elements."""

    def __init__(self, element_type: str, defaults: Defaults):
        self._type = element_type
        self._id = str(uuid.uuid4())
        self._seed = int(uuid.uuid4().int % 1000000000)
        self._version: int = 1
        self._version_nonce = int(uuid.uuid4().int % 1000000000)
        self._is_deleted = False
        self._x: float = 0
        self._y: float = 0
        self._width: float = getattr(defaults, "_width")
        self._height: float = getattr(defaults, "_height")
        self._opacity: int = getattr(defaults, "_opacity")
        self._angle: float = getattr(defaults, "_angle")
        self._index: str | None = None
        self._group_ids: list[str] = []
        self._frame_id: str | None = None
        self._link: None | str = None
        self._bound_elements = None
        self.__is_centered = False

    def position(self, x: float, y: float) -> Self:
        """
        Sets the position of the element.

        Args:
            x (float): The x-coordinate of the element.
            y (float): The y-coordinate of the element.

        Returns:
            Self: The instance of the element with updated position.
        """
        self._x = x
        self._y = y
        return self
    
    @overload
    def center(self) -> tuple[float, float]: ...

    @overload
    def center(self, x: float, y: float) -> Self: ...

    def center(self, *args) -> Self | tuple[float, float]:
        """
        Get or set the center coordinates of the element.

        Args:
            *args: Supports two signatures:
                1. center() -> tuple[float, float]
                    Returns the (x, y) coordinates of the center.
                2. center(x, y) -> Self
                    Sets the center to (x, y) and returns self for chaining.

        Returns:
            tuple[float, float] | Self: Depending on the arguments.
        """
        match args:
            case ():
                return (self._x + 0.5 * self._width, self._y + 0.5 * self._height)
            case (x, y):
                self.__is_centered = True
                self._x = x - 0.5 * self._width
                self._y = y - 0.5 * self._height
                return self
            case _:
                raise ValueError("Invalid arguments for center. Expected () or (x, y).")

    @overload
    def orbit(self, element: "AbstractElement", radius: float, angle: float) -> Self: ...

    @overload
    def orbit(self, x: float, y: float, radius: float, angle: float) -> Self: ...

    def orbit(self, *args) -> Self:
        """
        Positions the element relative to a reference using polar coordinates.
        
        This method allows placing the element such that its center will be at (radius, angle)
        from a reference. The reference can be either another AbstractElement or a point (x, y).

        Args:
            *args: Supports two signatures:
                1. orbit(element, radius, angle)
                    element (AbstractElement): The reference element to orbit around.
                    radius (float): The distance from the center of the reference.
                    angle (float): The angle to position the element at, in radians.
                
                2. orbit(x, y, radius, angle)
                    x (float): The x-coordinate of the reference point.
                    y (float): The y-coordinate of the reference point.
                    radius (float): The distance from the reference point.
                    angle (float): The angle to position the element at, in radians.

        Returns:
            Self: The instance of the element.
            
        Raises:
            ValueError: If the arguments do not match the expected signatures.
        """
        match args:
            case (AbstractElement() as ref, radius, angle):
                cx, cy = ref.center()
            case (cx, cy, radius, angle):
                pass
            case _:
                raise ValueError("Invalid arguments for orbit. Expected (element, radius, angle) or (x, y, radius, angle).")

        x = cx + radius * math.cos(angle)
        y = cy + radius * math.sin(angle)
        
        return self.center(x, y)

    def rotate(self, angle: float) -> Self:
        """
        Rotate the element clockwise by a specified angle.

        Args:
            angle (float): The angle to rotate the element clockwise by, in radians.

        Returns:
            Self: The instance of the element after rotation.
        """
        self._angle = angle
        return self

    def opacity(self, opacity: int) -> Self:
        """
        Set the opacity of the element.

        Args:
            opacity (int): The opacity value to set, must be in the range 0-100. 100 is fully opaque, 0 is fully transparent.

        Returns:
            Self: The instance of the element with updated opacity.

        Raises:
            ValueError: If the opacity value is not within the range 0-100.
        """
        self._opacity = Opacity.from_(opacity)
        return self

    def link(self, target: "str | AbstractElement") -> Self:
        """
        Establishes a link to the given target, which can be either a string URL or an AbstractElement instance.

        Args:
            target (str | AbstractElement): The target to link to. If it's an AbstractElement, a URL will be generated using its ID. 
                             If it's a string, it will be used directly as the link.

        Returns:
            Self: The instance of the current object with the updated link.

        Raises:
            ValueError: If the target is neither a string nor an AbstractElement.
        """
        match target:
            case AbstractElement():
                self._link = f"https://excalidraw.com/?element={target._id}"
            case str():
                self._link = target
            case _:
                raise ValueError("Link target must be a string or an AbstractElement.")
        return self

    @overload
    def size(self) -> tuple[float, float]: ...

    @overload
    def size(self, width: float, height: float) -> Self: ...

    def size(self, *args) -> Self | tuple[float, float]:
        """
        Get or set the size of the element.

        Args:
            *args: Supports two signatures:
                1. size() -> tuple[float, float]
                    Returns the (width, height) of the element.
                2. size(width, height) -> Self
                    Sets the size to (width, height) and returns self for chaining.

        Returns:
            tuple[float, float] | Self: Depending on the arguments.
        """
        match args:
            case ():
                return (self._width, self._height)
            case (width, height):
                return self._size(width, height)
            case _:
                raise ValueError("Invalid arguments for size. Expected () or (width, height).")

    def _size(self, width: float, height: float) -> Self:
        """Set the shape size."""
        if self.__is_centered:
            self._x = self._x + 0.5 * self._width - 0.5 * width
            self._y = self._y + 0.5 * self._height - 0.5 * height

        self._width = width
        self._height = height
        
        return self

    def _add_bound_element(self, element: "AbstractElement") -> None:
        self._bound_elements = self._bound_elements or []

        if not element._id in self._bound_elements: # TODO check if this works to prevent binding twice
            self._bound_elements.append({"id": element._id, "type": element._type})

    def _add_group_id(self, id: str) -> None:
        """Add a group ID to the element.

        Args:
            id (str): The group ID to add.
        """
        self._group_ids.append(id)



"""
Description: Base class for shapes with corners.
"""
# Copyright (c) 2024 - 2025 Milan Piskla
# Licensed under the MIT License - see LICENSE file for details


from .AbstractStrokedElement import AbstractStrokedElement
from .AbstractShape import AbstractShape
from .AbstractRoundableElement import AbstractRoundableElement
from .AbstractLabeledElement import AbstractLabeledElement
from ..base.AbstractPlainLabelListener import AbstractPlainLabelListener
from ..elements.Text import Text
from ...defaults.Defaults import Defaults
from typing import Self, Any

class AbstractCorneredShape(AbstractStrokedElement, AbstractShape, AbstractRoundableElement, AbstractLabeledElement):
    def __init__(self, type: str, defaults: Defaults, listener: AbstractPlainLabelListener, label: str | Text | None = None):
        super().__init__(type, defaults)
        self._init_labels(listener, label)
        self._roundness: str | dict[str, Any] | None = getattr(defaults, "_roundness")



"""
Description: Minimal Excaligen example.
"""
# Copyright (c) 2024 - 2026 Milan Piskla
# Licensed under the MIT License - see LICENSE file for details

from excaligen.SceneBuilder import SceneBuilder

scene = SceneBuilder()

scene.image().file("assets/robot.svg").center(0, 0)
scene.text("Oh look, I'm expressing joy").center(0, -150)
scene.text("how utterly revolting").center(0, 130)

scene.save("image.excalidraw")
